from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Skills


class FantasticVoyage(ElementalBurstBase):
    name: Literal['Fantastic Voyage'] = 'Fantastic Voyage'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 4,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_team_status('Inspiration Field', {
                'talent_activated': self.is_talent_equipped(match)
            })
        ]


# Talents


class GrandExpectation_3_3(SkillTalent):
    name: Literal['Grand Expectation']
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Bennett'] = 'Bennett'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 4,
        charge = 2
    )
    skill: Literal['Fantastic Voyage'] = 'Fantastic Voyage'


# charactor base


class Bennett_3_3(CharactorBase):
    name: Literal['Bennett']
    version: Literal['3.3'] = '3.3'
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | ElementalSkillBase | FantasticVoyage
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Strike of Fortune',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ElementalSkillBase(
                name = 'Passion Overload',
                damage_type = DamageElementalType.PYRO,
                cost = ElementalSkillBase.get_cost(self.element),
            ),
            FantasticVoyage(),
        ]


register_class(Bennett_3_3 | GrandExpectation_3_3)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import SummonBase

from ...modifiable_values import DamageValue
from ...event import RoundEndEventArguments

from ...action import (
    Actions, ChangeObjectUsageAction, MakeDamageAction, RemoveObjectAction
)
from ...struct import Cost

from ...consts import (
    DamageElementalType, DamageType, DieColor, ElementType, FactionType, 
    IconType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Charactor status. DO NOT define here, define in server/status/characor_status
# Here is just example.


# Summons


class LightfallSword_3_8(SummonBase):
    name: Literal['Lightfall Sword'] = 'Lightfall Sword'
    version: Literal['3.8'] = '3.8'
    usage: int = 0
    max_usage: int = 999
    damage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.PHYSICAL
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        When round end, make damage to the opponent and remove self
        """
        player_idx = self.position.player_idx
        target_table = match.player_tables[1 - player_idx]
        target_charactor = target_table.get_active_charactor()
        return [
            MakeDamageAction(
                damage_value_list = [
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.DAMAGE,
                        target_position = target_charactor.position,
                        damage = self.damage + self.usage,
                        damage_elemental_type = self.damage_elemental_type,
                        cost = Cost(),
                    )
                ],
            ),
            RemoveObjectAction(
                object_position = self.position,
            )
        ]


# Skills


class FavoniusBladeworkEdel(PhysicalNormalAttackBase):
    name: Literal['Favonius Bladework - Edel'] = 'Favonius Bladework - Edel'
    cost: Cost = PhysicalNormalAttackBase.get_cost(ElementType.CRYO)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If summon Lightfall Sword exists, not charge but add two usage.
        """
        summons = match.player_tables[self.position.player_idx].summons
        LS_idx = -1
        for sidx, summon in enumerate(summons):
            if summon.name == 'Lightfall Sword':
                LS_idx = sidx
                break
        ret: List[Actions] = []
        ret.append(self.attack_opposite_active(
            match, self.damage, self.damage_type))
        if LS_idx != -1:
            # add LS usage
            ret.append(ChangeObjectUsageAction(
                object_position = summons[LS_idx].position,
                change_usage = 2,
            ))
        else:
            # charge
            ret.append(self.charge_self(1))
        return ret


class IcetideVortex(ElementalSkillBase):
    name: Literal['Icetide Vortex'] = 'Icetide Vortex'
    version: Literal['3.8'] = '3.8'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If summon Lightfall Sword exists, not charge but add two or three 
        usage, based on whether has talent.
        """
        summons = match.player_tables[self.position.player_idx].summons
        LS_idx = -1
        for sidx, summon in enumerate(summons):
            if summon.name == 'Lightfall Sword':
                LS_idx = sidx
                break
        ret: List[Actions] = []
        ret.append(self.attack_opposite_active(
            match, self.damage, self.damage_type))
        if LS_idx != -1:
            # add LS usage
            usage = 2
            if self.is_talent_equipped(match):
                # has talent, add one more
                usage = 3
            ret.append(ChangeObjectUsageAction(
                object_position = summons[LS_idx].position,
                change_usage = usage,
            ))
        else:
            # charge
            ret.append(self.charge_self(1))
        status = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx].status
        for s in status:
            if s.name == 'Grimheart':
                # has Grimheart, return
                return ret
        # no Grimheart, add Grimheart
        ret.append(self.create_charactor_status(
            'Grimheart', { 'version': self.version }))
        return ret


class GlacialIllumination(ElementalBurstBase):
    name: Literal['Glacial Illumination'] = 'Glacial Illumination'
    version: Literal['3.8'] = '3.8'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and summon
        """
        return super().get_actions(match) + [
            self.create_summon('Lightfall Sword', { 'version': self.version })
        ]


# Talents


class WellspringOfWarLust_3_5(SkillTalent):
    name: Literal['Wellspring of War-Lust']
    version: Literal['3.5'] = '3.5'
    charactor_name: Literal['Eula'] = 'Eula'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3,
        charge = 2
    )
    skill: Literal['Glacial Illumination'] = 'Glacial Illumination'


# charactor base


class Eula_3_8(CharactorBase):
    name: Literal['Eula']
    version: Literal['3.8'] = '3.8'
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        FavoniusBladeworkEdel | IcetideVortex | GlacialIllumination
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            FavoniusBladeworkEdel(),
            IcetideVortex(),
            GlacialIllumination()
        ]


register_class(Eula_3_8 | WellspringOfWarLust_3_5 | LightfallSword_3_8)



import os
from src.lpsim import MatchState
from tests.utils_for_test import (
    get_pidx_cidx, get_test_id_from_command, make_respond, set_16_omni, 
    read_from_log_json
)


def test_new_4_tranform():
    match, agent_0, agent_1 = read_from_log_json(
        os.path.join(os.path.dirname(__file__), 
                     'jsons', 
                     'test_new_4_transform.json')
    )
    # add omnipotent guide
    set_16_omni(match)
    match.start()
    match.step()
    new_commands = [[], []]
    while True:
        if match.need_respond(0):
            agent = agent_0
            nc = new_commands[0]
        elif match.need_respond(1):
            agent = agent_1
            nc = new_commands[1]
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            nc.append(agent.commands[0])
            cmd = agent.commands[0].strip().split(' ')
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 10:
                pidx, cidx = get_pidx_cidx(cmd)
                desc = match.player_tables[pidx].charactors[cidx].desc
                cmd_desc = ' '.join(cmd[3:])[1:-1]
                assert desc == cmd_desc
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == '__main__':
    test_new_4_tranform()



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....consts import ObjectPositionType, SkillType, WeaponType

from ....action import ChargeAction
from ....event import SkillEndEventArguments
from ....struct import Cost
from .base import RoundEffectWeaponBase


class FavoniusBase(RoundEffectWeaponBase):
    name: str
    cost: Cost = Cost(same_dice_number = 3)
    version: str
    weapon_type: WeaponType
    max_usage_per_round: int = 1

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[ChargeAction]:
        """
        if self charactor use elemental skill, charge one more
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, 
            source_area = ObjectPositionType.CHARACTOR
        ):
            # not self charactor or not equipped
            return []
        if event.action.skill_type != SkillType.ELEMENTAL_SKILL:
            # not elemental skill
            return []
        if self.usage <= 0:
            # no usage
            return []
        self.usage -= 1
        return [ChargeAction(
            player_idx = self.position.player_idx,
            charactor_idx = self.position.charactor_idx,
            charge = 1
        )]


class FavoniusSword_3_7(FavoniusBase):
    name: Literal['Favonius Sword']
    version: Literal['3.7'] = '3.7'
    weapon_type: Literal[WeaponType.SWORD] = WeaponType.SWORD


register_class(FavoniusSword_3_7)



from typing import Any, List, Literal

from ....utils.class_registry import register_class
from ...struct import Cost

from ...action import (
    ActionTypes, Actions, CreateObjectAction, DrawCardAction, MakeDamageAction,
    RemoveObjectAction
)

from ...event import (
    MakeDamageEventArguments, RoundEndEventArguments, 
    RoundPrepareEventArguments, SkillEndEventArguments
)

from ...consts import (
    DamageElementalType, DamageType, DieColor, IconType, ObjectPositionType, 
    SkillType
)

from ...modifiable_values import (
    CostValue, DamageElementEnhanceValue, DamageIncreaseValue, DamageValue
)
from .base import (
    DefendCharactorStatus, ElementalInfusionCharactorStatus, 
    PrepareCharactorStatus, ReviveCharactorStatus, RoundCharactorStatus, 
    RoundEndAttackCharactorStatus, ShieldCharactorStatus, UsageCharactorStatus
)


class Stealth_3_3(DefendCharactorStatus, ElementalInfusionCharactorStatus):
    name: Literal['Stealth'] = 'Stealth'
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1
    infused_elemental_type: DamageElementalType = DamageElementalType.PYRO

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self, value: DamageElementEnhanceValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageElementEnhanceValue:
        """
        When self use skill, and has talent, change physical to pyro
        """
        assert mode == 'REAL'
        charactor = match.player_tables[
            self.position.player_idx].charactors[self.position.charactor_idx]
        if charactor.talent is None:
            # no talent, do nothing
            return value
        return super().value_modifier_DAMAGE_ELEMENT_ENHANCE(
            value, match, mode
        )

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        When self use skill, increase damage by 1
        """
        assert mode == 'REAL'
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not corresponding charactor use damage skill, do nothing
            return value
        if self.usage <= 0:  # pragma: no cover
            # no usage, do nothing
            return value
        # increase damage by 1
        value.damage += 1
        self.usage -= 1
        return value


class ExplosiveSpark_3_4(UsageCharactorStatus):
    name: Literal['Explosive Spark'] = 'Explosive Spark'
    version: Literal['3.4'] = '3.4'
    usage: int
    max_usage: int
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        assert mode == 'REAL'
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this charactor use normal attack, not modify
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charged attack, not modify
            return value
        # modify damage
        assert self.usage > 0
        self.usage -= 1
        value.damage += 1
        return value

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        If self use charged attack, decrease one 
        unaligned die cost.
        """
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True, 
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
        ):
            # not this charactor use skill, not modify
            return value
        skill = match.get_object(value.position)
        skill_type: SkillType = skill.skill_type
        if skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack, not modify
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charged attack, not modify
            return value
        # try decrease pyro cost
        value.cost.decrease_cost(DieColor.PYRO)
        return value


class NiwabiEnshou_3_3(ElementalInfusionCharactorStatus, UsageCharactorStatus):
    name: Literal['Niwabi Enshou'] = 'Niwabi Enshou'
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    infused_elemental_type: DamageElementalType = DamageElementalType.PYRO
    icon_type: Literal[IconType.ATK_UP_FIRE] = IconType.ATK_UP_FIRE

    effect_triggered: bool = False

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        If corresponding charactor use normal attack, increase damage by 1,
        and mark effect triggered.
        """
        assert mode == 'REAL'
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this charactor use normal attack, not modify
            return value
        # this charactor use normal attack, modify
        self.usage -= 1
        self.effect_triggered = True
        value.damage += 1
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        do not remove immediately here, as may trigger additional attack at
        skill end.
        """
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        If effect triggered, and have talent on yoimiya, make 1 pyro damage.
        If usage is zero, then remove self.
        """
        ret: List[MakeDamageAction | RemoveObjectAction] = []
        if self.position.player_idx != event.action.position.player_idx:
            # not self player use skill
            self.effect_triggered = False
            return list(self.check_should_remove())
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        if (
            charactor.name == 'Yoimiya'
            and event.action.skill_type == SkillType.NORMAL_ATTACK
            and self.effect_triggered
            and charactor.talent is not None
        ):
            # yoimiya use normal attack and self effect triggered
            # and has talent, attack 1 pyro damage
            target = match.player_tables[
                1 - self.position.player_idx].get_active_charactor()
            ret.append(MakeDamageAction(
                damage_value_list = [
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.DAMAGE,
                        target_position = target.position,
                        damage = 1,
                        damage_elemental_type = DamageElementalType.PYRO,
                        cost = Cost(),
                    )
                ]
            ))
        # reset mark
        self.effect_triggered = False
        return ret + self.check_should_remove()


class Brilliance_3_8(RoundCharactorStatus):
    name: Literal['Brilliance'] = 'Brilliance'
    version: Literal['3.8'] = '3.8'
    usage: int = 2
    max_usage: int = 2

    decrease_cost_usage: int = 1
    decrease_cost_max_usage: int = 1
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def renew(self, new_status: 'Brilliance_3_8') -> None:
        self.decrease_cost_usage = max(new_status.decrease_cost_usage, 
                                       self.decrease_cost_usage)
        self.decrease_cost_max_usage = max(new_status.decrease_cost_max_usage, 
                                           self.decrease_cost_max_usage)
        return super().renew(new_status)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        self.decrease_cost_usage = self.decrease_cost_max_usage
        return super().event_handler_ROUND_PREPARE(event, match)

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL'],
    ) -> CostValue:
        if self.decrease_cost_usage <= 0:
            # no usage left
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charge attack
            return value
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL
        ):
            # not self use skill
            return value
        skill = match.get_object(value.position)
        if skill.skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack
            return value
        # decrease cost
        if value.cost.decrease_cost(DieColor.PYRO):  # pragma: no branch
            if mode == 'REAL':
                self.decrease_cost_usage -= 1
        return value

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        return [CreateObjectAction(
            object_name = 'Scarlet Seal',
            object_position = self.position,
            object_arguments = {}
        )]


class ScarletSeal_4_2(UsageCharactorStatus):
    name: Literal['Scarlet Seal'] = 'Scarlet Seal'
    version: Literal['4.2'] = '4.2'
    usage: int = 1
    max_usage: int = 2
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP
    triggered: bool = False

    available_handler_in_trashbin: List[ActionTypes] = [
        ActionTypes.SKILL_END
    ]

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not self use normal attack
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charge attack
            return value
        # increase damage
        assert mode == 'REAL'
        self.usage -= 1
        self.triggered = True
        value.damage += 2
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[DrawCardAction]:
        """
        When triggered, and self charactor have talent, talent need to draw,
        then draw one card.
        """
        if not self.triggered:
            # not triggered
            return []
        self.triggered = False
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        if charactor.talent is not None and charactor.talent.draw_card:
            # have talent and need draw
            return [DrawCardAction(
                player_idx = self.position.player_idx,
                number = 1,
                draw_if_filtered_not_enough = True
            )]
        return []


class ParamitaPapilio_3_7(ElementalInfusionCharactorStatus, 
                          RoundCharactorStatus):
    name: Literal['Paramita Papilio'] = 'Paramita Papilio'
    version: Literal['3.7'] = '3.7'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    infused_elemental_type: DamageElementalType = DamageElementalType.PYRO

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not self use damage skill
            return value
        if (
            value.damage_elemental_type != DamageElementalType.PYRO
        ):  # pragma: no cover
            # not pyro damage
            return value
        # increase damage
        value.damage += 1
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If self use charged attack, create blood blossom on target
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL
        ):
            # not self use skill
            return []
        if not (
            event.action.skill_type == SkillType.NORMAL_ATTACK
            and match.player_tables[self.position.player_idx].charge_satisfied
        ):
            # not charged attack
            return []
        return [CreateObjectAction(
            object_position = event.action.target_position.set_area(
                ObjectPositionType.CHARACTOR_STATUS),
            object_name = 'Blood Blossom',
            object_arguments = {}
        )]


class BloodBlossom_3_7(RoundEndAttackCharactorStatus):
    name: Literal['Blood Blossom'] = 'Blood Blossom'
    version: Literal['3.7'] = '3.7'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.DOT] = IconType.DOT
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO


class DilucInfusion_3_3(ElementalInfusionCharactorStatus, 
                        RoundCharactorStatus):
    name: Literal['Pyro Elemental Infusion'] = 'Pyro Elemental Infusion'
    mark: Literal['Diluc'] = 'Diluc'
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2


class FieryRebirth_3_7(ReviveCharactorStatus):
    name: Literal['Fiery Rebirth'] = 'Fiery Rebirth'
    version: Literal['3.7'] = '3.7'
    heal: int = 3


class AegisOfAbyssalFlame_3_7(ShieldCharactorStatus):
    name: Literal['Aegis of Abyssal Flame'] = 'Aegis of Abyssal Flame'
    version: Literal['3.7'] = '3.7'
    usage: int = 3
    max_usage: int = 3

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, 
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        if (
            value.is_corresponding_charactor_use_damage_skill(
                self.position, match, None
            )
            and value.damage_elemental_type == DamageElementalType.PYRO
        ):
            # self attack and pyro damage, increase 1 damage
            assert mode == 'REAL'
            value.damage += 1
        return value


class IncinerationDrive_4_1(PrepareCharactorStatus):
    name: Literal['Incineration Drive'] = 'Incineration Drive'
    version: Literal['4.1'] = '4.1'
    charactor_name: Literal['Dehya'] = 'Dehya'
    skill_name: Literal['Incineration Drive'] = 'Incineration Drive'


class ScarletSeal_3_8(ScarletSeal_4_2):
    version: Literal['3.8']
    usage: int = 1
    max_usage: int = 1


register_class(
    Stealth_3_3 | ExplosiveSpark_3_4 | NiwabiEnshou_3_3 | Brilliance_3_8 
    | ScarletSeal_4_2 | ParamitaPapilio_3_7 | BloodBlossom_3_7 
    | DilucInfusion_3_3 | FieryRebirth_3_7 | AegisOfAbyssalFlame_3_7 
    | IncinerationDrive_4_1 | ScarletSeal_3_8
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions
from ...consts import DamageElementalType
from .xingqiu_3_6 import Xingqiu_3_6 as X_3_6
from .xingqiu_3_6 import Raincutter as R_3_6
from .xingqiu_4_1 import GuhuaStyle, FatalRainscreen


class Raincutter(R_3_6):
    name: Literal['Raincutter'] = 'Raincutter'

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack, application and create version 3.3 object
        """
        ret = [
            self.charge_self(-2),
            self.attack_opposite_active(match, self.damage, self.damage_type),
            self.element_application_self(match, DamageElementalType.HYDRO)
        ]
        return ret + [
            self.create_team_status('Rainbow Bladework', {
                'version': '3.3'
            })
        ]


class Xingqiu_3_3(X_3_6):
    version: Literal['3.3']
    skills: List[GuhuaStyle | FatalRainscreen | Raincutter] = []

    def _init_skills(self) -> None:
        self.skills = [
            GuhuaStyle(),
            FatalRainscreen(),
            Raincutter()
        ]


register_class(Xingqiu_3_3)



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....modifiable_values import DamageValue

from ....consts import (
    DamageElementalType, DamageType, ObjectPositionType, SkillType
)

from ....action import MakeDamageAction

from ....event import SkillEndEventArguments

from ....struct import Cost
from .base import RoundEffectArtifactBase


class HealBySkillArtifactBase(RoundEffectArtifactBase):
    name: str
    version: str
    cost: Cost
    max_usage_per_round: int
    heal: int
    heal_skill_type: SkillType
    heal_target: Literal['SELF', 'TEAM']

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If self use elemental skill, create one corresponding dice
        """
        if not (
            self.position.area == ObjectPositionType.CHARACTOR
            and event.action.position.player_idx == self.position.player_idx 
            and event.action.position.charactor_idx 
            == self.position.charactor_idx
            and event.action.skill_type == self.heal_skill_type
            and self.usage > 0
        ):
            # not equipped or not self use elemental skill or no usage
            return []
        self.usage -= 1
        action = MakeDamageAction(
            damage_value_list = [],
        )
        if self.heal_target == 'SELF':
            charactor = match.player_tables[
                self.position.player_idx].charactors[
                    self.position.charactor_idx]
            action.damage_value_list.append(
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = charactor.position,
                    damage = -self.heal,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = self.cost.copy()
                )
            )
        else:
            charactors = match.player_tables[
                self.position.player_idx].charactors
            for charactor in charactors:
                if charactor.is_defeated:
                    continue
                action.damage_value_list.append(
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.HEAL,
                        target_position = charactor.position,
                        damage = -self.heal,
                        damage_elemental_type = DamageElementalType.HEAL,
                        cost = self.cost.copy()
                    )
                )
        return [action]


class AdventurersBandana_3_3(HealBySkillArtifactBase):
    name: Literal["Adventurer's Bandana"] = "Adventurer's Bandana"
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 1)
    max_usage_per_round: int = 3
    heal_skill_type: SkillType = SkillType.NORMAL_ATTACK
    heal: int = 1
    heal_target: Literal['SELF', 'TEAM'] = 'SELF'


class LuckyDogsSilverCirclet_3_3(HealBySkillArtifactBase):
    name: Literal["Lucky Dog's Silver Circlet"]
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(any_dice_number = 2)
    max_usage_per_round: int = 1
    heal_skill_type: SkillType = SkillType.ELEMENTAL_SKILL
    heal: int = 2
    heal_target: Literal['SELF', 'TEAM'] = 'SELF'


class TravelingDoctorsHandkerchief_3_3(HealBySkillArtifactBase):
    name: Literal["Traveling Doctor's Handkerchief"]
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 1)
    max_usage_per_round: int = 1
    heal_skill_type: SkillType = SkillType.ELEMENTAL_BURST
    heal: int = 1
    heal_target: Literal['SELF', 'TEAM'] = 'TEAM'


register_class(
    AdventurersBandana_3_3 | LuckyDogsSilverCirclet_3_3 
    | TravelingDoctorsHandkerchief_3_3
)



from typing import List, Tuple, Literal

from .charactor.charactor_base import CharactorBase
from .consts import (
    ElementType, ElementalReactionType, DamageElementalType,
    DamageType, ELEMENT_TO_DAMAGE_TYPE, ObjectPositionType,
)
from .modifiable_values import DamageElementEnhanceValue, DamageIncreaseValue
from .action import Actions, CreateObjectAction
from .struct import ObjectPosition


def check_elemental_reaction(
    source: ElementType, targets: List[ElementType]
) -> Tuple[ElementalReactionType, List[ElementType], 
           List[ElementType]]:
    """
    Check the elemental reaction based on source and targets.

    Args:
        source (ElementType): The element type of attack source.
        targets (List[ElementType]): The element application of attack targets.

    Returns:
        Tuple[ElementalReactionType, List[ElementType], List[ElementType]]:
            The elemental reaction type, the element types that are consumed,
            and the element applications updated.
    """
    assert len(targets) <= 2, \
        f'Two element application at most, got {targets}.'
    if len(targets) == 2:
        assert targets == [ElementType.CRYO, ElementType.DENDRO], \
            'Two element application must be cryo and dendro.' \
            f'Got {targets}.'
    assert ElementType.GEO not in targets, \
        'Geo cannot be applied to enemies.'
    assert ElementType.ANEMO not in targets, \
        'Anemo cannot be applied to enemies.'
    if source == ElementType.NONE:
        return ElementalReactionType.NONE, [], targets
    elif source == ElementType.CRYO:
        if ElementType.PYRO in targets:
            return (
                ElementalReactionType.MELT, 
                [ElementType.PYRO, ElementType.CRYO], 
                []
            )
        elif ElementType.HYDRO in targets:
            return (
                ElementalReactionType.FROZEN, 
                [ElementType.HYDRO, ElementType.CRYO], 
                []
            )
        elif ElementType.ELECTRO in targets:
            return (
                ElementalReactionType.SUPERCONDUCT, 
                [ElementType.ELECTRO, ElementType.CRYO], 
                []
            )
        else:
            new_targets = [ElementType.CRYO]
            if ElementType.DENDRO in targets:
                new_targets.append(ElementType.DENDRO)
            return (
                ElementalReactionType.NONE, 
                [], 
                new_targets
            )
    elif source == ElementType.HYDRO:
        if ElementType.PYRO in targets:
            return (
                ElementalReactionType.VAPORIZE, 
                [ElementType.HYDRO, ElementType.PYRO], 
                []
            )
        elif ElementType.CRYO in targets:
            return (
                ElementalReactionType.FROZEN, 
                [ElementType.HYDRO, ElementType.CRYO], 
                targets[1:]  # cryo must be first
            )
        elif ElementType.ELECTRO in targets:
            return (
                ElementalReactionType.ELECTROCHARGED, 
                [ElementType.ELECTRO, ElementType.HYDRO], 
                []
            )
        elif ElementType.DENDRO in targets:
            return (
                ElementalReactionType.BLOOM, 
                [ElementType.HYDRO, ElementType.DENDRO], 
                []
            )
        else:
            return (
                ElementalReactionType.NONE, 
                [], 
                [ElementType.HYDRO]
            )
    elif source == ElementType.PYRO:
        if ElementType.HYDRO in targets:
            return (
                ElementalReactionType.VAPORIZE, 
                [ElementType.HYDRO, ElementType.PYRO], 
                []
            )
        elif ElementType.CRYO in targets:
            return (
                ElementalReactionType.MELT, 
                [ElementType.PYRO, ElementType.CRYO], 
                targets[1:]  # cryo must be first
            )
        elif ElementType.ELECTRO in targets:
            return (
                ElementalReactionType.OVERLOADED, 
                [ElementType.PYRO, ElementType.ELECTRO], 
                []
            )
        elif ElementType.DENDRO in targets:
            return (
                ElementalReactionType.BURNING, 
                [ElementType.PYRO, ElementType.DENDRO], 
                []
            )
        else:
            return (
                ElementalReactionType.NONE, 
                [], 
                [ElementType.PYRO]
            )
    elif source == ElementType.ELECTRO:
        if ElementType.HYDRO in targets:
            return (
                ElementalReactionType.ELECTROCHARGED, 
                [ElementType.ELECTRO, ElementType.HYDRO], 
                []
            )
        elif ElementType.CRYO in targets:
            return (
                ElementalReactionType.SUPERCONDUCT, 
                [ElementType.ELECTRO, ElementType.CRYO], 
                targets[1:]  # cryo must be first
            )
        elif ElementType.PYRO in targets:
            return (
                ElementalReactionType.OVERLOADED, 
                [ElementType.PYRO, ElementType.ELECTRO], 
                []
            )
        elif ElementType.DENDRO in targets:
            return (
                ElementalReactionType.QUICKEN, 
                [ElementType.ELECTRO, ElementType.DENDRO], 
                []
            )
        else:
            return (
                ElementalReactionType.NONE, 
                [], 
                [ElementType.ELECTRO]
            )
    elif source == ElementType.GEO:
        if ElementType.HYDRO in targets:
            return (
                ElementalReactionType.CRYSTALLIZE, 
                [ElementType.GEO, ElementType.HYDRO], 
                []
            )
        elif ElementType.PYRO in targets:
            return (
                ElementalReactionType.CRYSTALLIZE, 
                [ElementType.GEO, ElementType.PYRO], 
                []
            )
        elif ElementType.CRYO in targets:
            return (
                ElementalReactionType.CRYSTALLIZE, 
                [ElementType.GEO, ElementType.CRYO], 
                targets[1:]  # cryo must be first
            )
        elif ElementType.ELECTRO in targets:
            return (
                ElementalReactionType.CRYSTALLIZE, 
                [ElementType.GEO, ElementType.ELECTRO], 
                []
            )
        else:
            return (
                ElementalReactionType.NONE, 
                [], 
                targets
            )
    elif source == ElementType.DENDRO:
        if ElementType.HYDRO in targets:
            return (
                ElementalReactionType.BLOOM, 
                [ElementType.HYDRO, ElementType.DENDRO], 
                []
            )
        elif ElementType.PYRO in targets:
            return (
                ElementalReactionType.BURNING, 
                [ElementType.PYRO, ElementType.DENDRO], 
                []
            )
        elif ElementType.ELECTRO in targets:
            return (
                ElementalReactionType.QUICKEN, 
                [ElementType.ELECTRO, ElementType.DENDRO], 
                []
            )
        else:
            new_targets = [ElementType.DENDRO]
            if ElementType.CRYO in targets:
                new_targets = [ElementType.CRYO, ElementType.DENDRO]
            return (
                ElementalReactionType.NONE, 
                [], 
                new_targets
            )
    else:
        assert source == ElementType.ANEMO
        if ElementType.HYDRO in targets:
            return (
                ElementalReactionType.SWIRL, 
                [ElementType.ANEMO, ElementType.HYDRO], 
                []
            )
        elif ElementType.PYRO in targets:
            return (
                ElementalReactionType.SWIRL, 
                [ElementType.ANEMO, ElementType.PYRO], 
                []
            )
        elif ElementType.CRYO in targets:
            return (
                ElementalReactionType.SWIRL, 
                [ElementType.ANEMO, ElementType.CRYO], 
                targets[1:]  # cryo must be first
            )
        elif ElementType.ELECTRO in targets:
            return (
                ElementalReactionType.SWIRL, 
                [ElementType.ANEMO, ElementType.ELECTRO], 
                []
            )
        else:
            return (
                ElementalReactionType.NONE, 
                [], 
                targets
            )


def apply_elemental_reaction(
        target_charactors: List[CharactorBase],
        active_charactor_idx: int,
        damage: DamageElementEnhanceValue,
        reaction: ElementalReactionType,
        reacted_elements: List[ElementType]
) -> Tuple[DamageIncreaseValue, List[DamageElementEnhanceValue]]:
    """
    Apply elemental reaction to DamageElementEnhanceValue, and
    generate DamageIncreaseValue.
    It returns a DamageIncreaseValue related to input damage, and a list of
    DamageElementEnhanceValue that generated by elemental reaction.

    NOTE: Only direct damage will count, summons and status will not 
    considered to generate here.
    """
    res: List[DamageElementEnhanceValue] = []
    damage = DamageIncreaseValue.from_element_enhance_value(
        damage, reaction, reacted_elements
    )
    if reaction == ElementalReactionType.NONE:
        return damage, res
    if damage.damage_type != DamageType.DAMAGE:
        return damage, res
    if reaction in [ElementalReactionType.MELT,
                    ElementalReactionType.VAPORIZE,
                    ElementalReactionType.OVERLOADED]:
        # +2 damage types
        damage.damage += 2
    elif reaction == ElementalReactionType.SWIRL:
        # special swirl, 1 elemental damage for other charactors
        element_type = reacted_elements[1]
        assert element_type != ElementType.ANEMO, \
            'Anemo should always be first'
        attack_target = damage.target_position.charactor_idx
        for cnum in range(0, len(target_charactors)):
            cnum = (cnum + active_charactor_idx) % len(target_charactors)
            if cnum == attack_target:
                continue
            c = target_charactors[cnum]
            if c.is_alive:
                res.append(DamageElementEnhanceValue(
                    position = damage.position,
                    target_position = ObjectPosition(
                        player_idx = damage.target_position.player_idx,
                        charactor_idx = cnum,
                        area = damage.target_position.area,
                        id = c.id,
                    ),
                    damage = 1,
                    damage_type = DamageType.DAMAGE,
                    damage_elemental_type = ELEMENT_TO_DAMAGE_TYPE[
                        element_type],
                    cost = damage.cost.copy(),
                    damage_from_element_reaction = True
                ))
    else:
        # all others are +1 damage types
        damage.damage += 1
        if reaction in [ElementalReactionType.ELECTROCHARGED,
                        ElementalReactionType.SUPERCONDUCT]:
            # 1 piercing dmg for other charactors
            damage_type = DamageElementalType.PIERCING
            attack_target = damage.target_position.charactor_idx
            for cnum in range(0, len(target_charactors)):
                cnum = (cnum + active_charactor_idx) % len(target_charactors)
                if cnum == attack_target:
                    continue
                c = target_charactors[cnum]
                if c.is_alive:
                    res.append(DamageElementEnhanceValue(
                        position = damage.position,
                        target_position = ObjectPosition(
                            player_idx = damage.target_position.player_idx,
                            charactor_idx = cnum,
                            area = damage.target_position.area,
                            id = c.id,
                        ),
                        damage = 1,
                        damage_type = DamageType.DAMAGE,
                        damage_elemental_type = damage_type,
                        cost = damage.cost.copy(),
                        damage_from_element_reaction = True
                    ))
    return damage, res


def elemental_reaction_side_effect_ver_3_4(
        reaction: ElementalReactionType, player_idx: int,
        charator_idx: int) -> Actions | None:
    """
    Apply side effect of elemental reaction. 
    """
    if reaction == ElementalReactionType.FROZEN:
        position = ObjectPosition(
            player_idx = player_idx,
            charactor_idx = charator_idx,
            area = ObjectPositionType.CHARACTOR_STATUS,
            id = -1,
        )
        return CreateObjectAction(
            object_position = position,
            object_name = 'Frozen',
            object_arguments = {}
        )
    elif reaction == ElementalReactionType.CRYSTALLIZE:
        position = ObjectPosition(
            player_idx = 1 - player_idx,
            area = ObjectPositionType.TEAM_STATUS,
            id = -1,
        )
        return CreateObjectAction(
            object_position = position,
            object_name = 'Crystallize',
            object_arguments = {}
        )
    elif reaction == ElementalReactionType.BURNING:
        position = ObjectPosition(
            player_idx = 1 - player_idx,
            area = ObjectPositionType.SUMMON,
            id = -1,
        )
        return CreateObjectAction(
            object_position = position,
            object_name = 'Burning Flame',
            object_arguments = {}
        )
    elif reaction == ElementalReactionType.BLOOM:
        position = ObjectPosition(
            player_idx = 1 - player_idx,
            area = ObjectPositionType.TEAM_STATUS,
            id = -1,
        )
        return CreateObjectAction(
            object_position = position,
            object_name = 'Dendro Core',
            object_arguments = {}
        )
    elif reaction == ElementalReactionType.QUICKEN:
        position = ObjectPosition(
            player_idx = 1 - player_idx,
            area = ObjectPositionType.TEAM_STATUS,
            id = -1,
        )
        return CreateObjectAction(
            object_position = position,
            object_name = 'Catalyzing Field',
            object_arguments = {}
        )
    return None


def elemental_reaction_side_effect_ver_3_3(
        reaction: ElementalReactionType, player_idx: int,
        charator_idx: int) -> Actions | None:
    """
    Apply side effect of elemental reaction. In 3.3, only quicken is different.
    """
    if reaction == ElementalReactionType.QUICKEN:
        position = ObjectPosition(
            player_idx = 1 - player_idx,
            area = ObjectPositionType.TEAM_STATUS,
            id = -1,
        )
        return CreateObjectAction(
            object_position = position,
            object_name = 'Catalyzing Field',
            object_arguments = { 'version': '3.3' }
        )
    return elemental_reaction_side_effect_ver_3_4(
        reaction, player_idx, charator_idx
    )


def elemental_reaction_side_effect(
        reaction: ElementalReactionType, player_idx: int,
        charator_idx: int, version: Literal['3.3', '3.4'] = '3.4'
) -> Actions | None:
    if version == '3.3':
        return elemental_reaction_side_effect_ver_3_3(
            reaction, player_idx, charator_idx
        )
    return elemental_reaction_side_effect_ver_3_4(
        reaction, player_idx, charator_idx
    )



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...struct import Cost

from ...modifiable_values import (
    DamageDecreaseValue, DamageIncreaseValue, DamageValue
)

from ...consts import (
    DamageElementalType, DamageType, IconType, ObjectPositionType, SkillType
)

from ...action import MakeDamageAction, RemoveObjectAction

from ...event import MakeDamageEventArguments, SkillEndEventArguments
from .base import (
    DefendTeamStatus, ExtraAttackTeamStatus, RoundTeamStatus, UsageTeamStatus
)


class SparksNSplash_3_4(UsageTeamStatus):
    name: Literal["Sparks 'n' Splash"] = "Sparks 'n' Splash"
    version: Literal['3.4'] = '3.4'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        When attached charactor use skill, then damage itself.
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True, 
            target_area = ObjectPositionType.SKILL,
        ):
            # not charactor use skill, not modify
            return []
        active_charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        if self.usage > 0:  # pragma: no branch
            self.usage -= 1
            return [MakeDamageAction(
                damage_value_list = [DamageValue(
                    position = self.position,
                    damage_type = DamageType.DAMAGE,
                    target_position = active_charactor.position,
                    damage = 2,
                    damage_elemental_type = DamageElementalType.PYRO,
                    cost = Cost()
                )]
            )]
        else:
            return []  # pragma: no cover


class InspirationField_3_3(RoundTeamStatus):
    name: Literal['Inspiration Field'] = 'Inspiration Field'
    desc: Literal['', 'talent'] = ''
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    talent_activated: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.talent_activated:
            self.desc = 'talent'

    def renew(self, new_status: 'InspirationField_3_3') -> None:
        super().renew(new_status)
        if new_status.talent_activated:
            self.talent_activated = True
            self.desc = 'talent'

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        If charactor HP greater than 7, or talent activated, increase damage
        by 2
        """
        assert mode == 'REAL'
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not our charactor use skill, do nothing
            return value
        charactor = match.player_tables[value.position.player_idx].charactors[
            value.position.charactor_idx]
        if not self.talent_activated and charactor.hp < 7:
            # no talent and hp less than 7, do nothing
            return value
        # increase damage by 2
        value.damage += 2
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If our charactor use skill and HP less than 7, heal 2 HP
        """
        if self.position.player_idx != event.action.position.player_idx:
            # not our charactor, do nothing
            return []
        charactor = match.player_tables[
            event.action.position.player_idx].charactors[
                event.action.position.charactor_idx]
        if charactor.hp > 6:
            # HP greater than 6, do nothing
            return []
        # heal this charactor by 2
        return [MakeDamageAction(
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = charactor.position,
                    damage = -2,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = Cost()
                )
            ],
        )]


class AurousBlaze_3_3(RoundTeamStatus, ExtraAttackTeamStatus):
    name: Literal['Aurous Blaze'] = 'Aurous Blaze'
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    trigger_skill_type: SkillType | None = None
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    decrease_usage: bool = False

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        if skill used by self player but not yoimiya, make damage to opponent
        active charactor.
        """
        charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        if charactor.name == 'Yoimiya':
            # yoimiya use skill
            return []
        return super().event_handler_SKILL_END(event, match)


class Pyronado_3_3(UsageTeamStatus, ExtraAttackTeamStatus):
    name: Literal['Pyronado'] = 'Pyronado'
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    newly_created: bool = True
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    trigger_skill_type: SkillType | None = None
    damage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    decrease_usage: bool = True

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If newly created, should not trigger immediately.
        """
        if self.newly_created:
            self.newly_created = False
            return []
        return super().event_handler_SKILL_END(event, match)


class FierySanctumField_4_1(DefendTeamStatus):
    name: Literal['Fiery Sanctum Field'] = 'Fiery Sanctum Field'
    version: Literal['4.1'] = '4.1'
    usage: int = 1
    max_usage: int = 1
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1
    decrease_usage_by_damage: bool = False
    remove_triggered: bool = False

    def _find_dehya(self, match: Any) -> int:
        """
        Find first alive and standby dehya. If not found, return -1.
        """
        charactors = match.player_tables[self.position.player_idx].charactors
        active_idx = match.player_tables[
            self.position.player_idx].active_charactor_idx
        for cidx, charactor in enumerate(charactors):
            if (
                charactor.name == 'Dehya'
                and charactor.is_alive
                and cidx != active_idx
            ):
                return cidx
        return -1

    def value_modifier_DAMAGE_DECREASE(
        self, value: DamageDecreaseValue, match: Any, 
        mode: Literal['TEST', 'REAL']
    ) -> DamageDecreaseValue:
        """
        Check if have alive dehya on standby, if not, no effect.
        """
        dehya_idx = self._find_dehya(match)
        if dehya_idx == -1:
            # not found dehya
            return value
        return super().value_modifier_DAMAGE_DECREASE(value, match, mode)

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        If self.shield_triggered, then check whether need to deal 1 piercing
        damage to dehya.
        """
        ret: List[MakeDamageAction | RemoveObjectAction] = []
        if self.usage == 0:
            if self.remove_triggered:
                # has triggered remove, do nothing
                return []
            self.remove_triggered = True
            # check if should attack dehya
            dehya_idx = self._find_dehya(match)
            assert dehya_idx != -1
            dehya = match.player_tables[
                self.position.player_idx].charactors[dehya_idx]
            if (
                dehya.hp >= 7
                and dehya.is_alive
            ):
                # after shield triggered, dehya has at least 7 hp and alive,
                # make 1 piercing damage to dehya
                ret.append(MakeDamageAction(
                    damage_value_list = [
                        DamageValue(
                            position = self.position,
                            target_position = dehya.position,
                            damage = 1,
                            damage_type = DamageType.DAMAGE,
                            damage_elemental_type 
                            = DamageElementalType.PIERCING,
                            cost = Cost(),
                        )
                    ]
                ))
        return ret + super().event_handler_MAKE_DAMAGE(event, match)


register_class(
    SparksNSplash_3_4 | InspirationField_3_3 | AurousBlaze_3_3 | Pyronado_3_3 
    | FierySanctumField_4_1
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import RemoveObjectAction

from ...event import MoveObjectEventArguments, UseSkillEventArguments

from ...consts import CostLabels, IconType, ObjectPositionType

from ...modifiable_values import CostValue
from .base import RoundCharactorStatus, ShieldCharactorStatus


class LithicSpear_3_3(ShieldCharactorStatus):
    name: Literal['Lithic Spear'] = 'Lithic Spear'
    version: Literal['3.3'] = '3.3'
    usage: int
    max_usage: int

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def renew(self, object: 'LithicSpear_3_3') -> None:
        self.max_usage = object.max_usage
        self.usage = max(object.usage, self.usage)
        self.usage = min(self.max_usage, self.usage)


class KingsSquire_4_0(RoundCharactorStatus):
    name: Literal["King's Squire"] = "King's Squire"
    version: Literal['4.0'] = '4.0'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        If self use elemental skill, or equip talent, cost -2.
        """
        label = value.cost.label
        if label & (
            CostLabels.ELEMENTAL_SKILL.value | CostLabels.TALENT.value
        ) == 0:  # no label match
            return value
        position = value.position
        if position.player_idx != self.position.player_idx:
            # not same player
            return value
        assert self.position.charactor_idx != -1
        if position.area == ObjectPositionType.SKILL:
            # cost from charactor
            if position.charactor_idx != self.position.charactor_idx:
                # not same charactor
                return value
        else:
            # cost from hand card, is a talent card
            charactor = match.player_tables[
                self.position.player_idx
            ].charactors[self.position.charactor_idx]
            for card in match.player_tables[
                    self.position.player_idx].hands:
                if card.id == value.position.id:
                    if card.charactor_name != charactor.name:
                        # talent card not for this charactor
                        return value
        # decrease cost
        assert self.usage > 0
        decrease_result = [
            value.cost.decrease_cost(value.cost.elemental_dice_color),
            value.cost.decrease_cost(value.cost.elemental_dice_color),
        ]
        if True in decrease_result:  # pragma: no branch
            # decrease cost success
            if mode == 'REAL':
                self.usage -= 1
        return value

    def event_handler_USE_SKILL(
        self, event: UseSkillEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        return self.check_should_remove()

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        return self.check_should_remove()


class Moonpiercer_4_1(KingsSquire_4_0):
    name: Literal['Moonpiercer'] = 'Moonpiercer'
    version: Literal['4.1'] = '4.1'


register_class(LithicSpear_3_3 | KingsSquire_4_0 | Moonpiercer_4_1)



from src.lpsim.agents import InteractionAgent
from src.lpsim import Deck, Match, MatchState
from tests.utils_for_test import get_random_state, make_respond, set_16_omni


def test_memento_lens():
    cmd_records = [
        [
            "sw_card",
            "choose 1",
            "sw_char 2 15",
            "card 1 0 14",
            "card 1 1 13 12",
            "sw_char 0 11",
            "end",
            "card 2 2",
            "card 2 2 15 14",
            "end",
            "card 1 0 15 14",
            "card 3 0 13 12",
            "card 1 0 11 10 9",
            "card 0 0 8",
            "card 0 0",
            "end",
            "sw_char 1 15"
        ],
        [
            "sw_card",
            "choose 1",
            "card 0 1 15 14",
            "sw_char 2 13",
            "card 1 2 12 11",
            "end",
            "card 3 0 15 14 13",
            "card 1 0 12",
            "sw_char 1 11",
            "end",
            "card 3 0 15",
            "card 2 0",
            "card 2 0 14",
            "card 0 0 13",
            "end",
            "card 0 0",
            "card 1 0 15 14 13",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        charactor:Ningguang@3.3
        charactor:Kaeya@3.3
        charactor:Chongyun@3.3
        Sacrificial Fragments@3.3*5
        Instructor's Cap@3.3*5
        Chinju Forest@3.7*5
        Paimon@3.3*5
        Memento Lens@4.3*5
        Abyssal Summons@3.3*5
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == '__main__':
    test_memento_lens()



from typing import Any, List, Literal, Tuple

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...event import (
    ChooseCharactorEventArguments, RoundPrepareEventArguments, 
    SwitchCharactorEventArguments
)

from ...action import (
    Actions, CreateObjectAction, RemoveObjectAction
)
from ...struct import Cost, ObjectPosition

from ...consts import (
    CostLabels, DamageElementalType, DieColor, ElementType, FactionType, 
    ObjectPositionType, PlayerActionLabels, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    CharactorBase, TalentBase
)


# Summons


class LightningRoseSummon_4_0(AttackerSummonBase):
    name: Literal['Lightning Rose'] = 'Lightning Rose'
    version: Literal['4.0'] = '4.0'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 2


# Skills


class LightningTouch(ElementalNormalAttackBase):
    name: Literal['Lightning Touch'] = 'Lightning Touch'
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = ElementalNormalAttackBase.get_cost(ElementType.ELECTRO)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Do attack; if charged attack, attach conductive
        """
        ret = super().get_actions(match)
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charged attack
            return ret
        ret.append(self.create_opposite_charactor_status(
            match, 'Conductive', {}
        ))
        return ret


class VioletArc(ElementalSkillBase):
    name: Literal['Violet Arc'] = 'Violet Arc'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If Conductive not attached, attach Conductive
        otherwise, increase damage.
        """
        target = match.player_tables[
            1 - self.position.player_idx].get_active_charactor()
        status = target.status
        conductive = None
        for s in status:
            if s.name == 'Conductive':
                conductive = s
                break
        if conductive is None:
            # no conductive
            return super().get_actions(match) + [
                self.create_opposite_charactor_status(
                    match, 'Conductive', {}
                )
            ]
        else:
            # has conductive
            self.damage += conductive.usage
            ret = super().get_actions(match)
            self.damage = 2
            ret = [RemoveObjectAction(
                object_position = conductive.position
            )] + ret
            return ret


class LightningRose(ElementalBurstBase):
    name: Literal['Lightning Rose'] = 'Lightning Rose'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_summon(self.name)
        ]


# Talents


class PulsatingWitch_4_0(TalentBase):
    name: Literal['Pulsating Witch']
    version: Literal['4.0'] = '4.0'
    charactor_name: Literal['Lisa'] = 'Lisa'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 1
    )
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.TALENT.value 
        | CostLabels.EQUIPMENT.value
    )
    usage: int = 1

    def is_valid(self, match: Any) -> bool:
        """
        Only corresponding charactor is active charactor can equip this card.
        """
        if self.position.area != ObjectPositionType.HAND:
            # not in hand, cannot equip
            raise AssertionError('Talent is not in hand')
        return len(self.get_targets(match)) > 0

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        All ayaka are targets
        """
        assert self.position.area != ObjectPositionType.CHARACTOR
        ret: List[ObjectPosition] = []
        for c in match.player_tables[self.position.player_idx].charactors:
            if c.name == self.charactor_name and c.is_alive:
                ret.append(c.position)
        return ret

    def get_action_type(self, match: Any) -> Tuple[int, bool]:
        return (
            PlayerActionLabels.CARD.value | PlayerActionLabels.SKILL.value,
            False
        )

    def _attach_conductive_to_opposite_active(
        self, 
        event: SwitchCharactorEventArguments | ChooseCharactorEventArguments,
        match: Any
    ) -> List[CreateObjectAction]:
        if (
            event.action.player_idx != self.position.player_idx
            or event.action.charactor_idx != self.position.charactor_idx
        ):
            # not to this charactor
            return []
        target = match.player_tables[
            1 - self.position.player_idx].get_active_charactor()
        if self.usage <= 0:
            # no usage
            return []
        self.usage -= 1
        return [CreateObjectAction(
            object_name = 'Conductive',
            object_position = target.position.set_area(
                ObjectPositionType.CHARACTOR_STATUS
            ),
            object_arguments = {}
        )]

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        Reset usage
        """
        self.usage = 1
        return []

    def event_handler_SWITCH_CHARACTOR(
        self, event: SwitchCharactorEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        return self._attach_conductive_to_opposite_active(event, match)

    def event_handler_CHOOSE_CHARACTOR(
        self, event: ChooseCharactorEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        return self._attach_conductive_to_opposite_active(event, match)


# charactor base


class Lisa_4_0(CharactorBase):
    name: Literal['Lisa']
    version: Literal['4.0'] = '4.0'
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        LightningTouch | VioletArc | LightningRose
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            LightningTouch(),
            VioletArc(),
            LightningRose()
        ]


register_class(Lisa_4_0 | PulsatingWitch_4_0 | LightningRoseSummon_4_0)



# type: ignore


from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...modifiable_values import CostValue
from ...event import ReceiveDamageEventArguments, RoundPrepareEventArguments

from ...action import Actions, MakeDamageAction
from ...struct import Cost

from ...consts import (
    CostLabels, DamageElementalType, DieColor, ElementType, FactionType, 
    ObjectPositionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Charactor status. DO NOT define here, define in server/status/characor_status
# Here is just example.
# Round status, will last for several rounds and disappear
# Skills


class Tidecaller(ElementalSkillBase):
    name: Literal['Tidecaller'] = 'Tidecaller'
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create status
        """
        return [
            self.charge_self(1),
            self.create_charactor_status('Tidecaller: Surf Embrace'),
        ]


class Wavestrider(ElementalSkillBase):
    name: Literal['Wavestrider'] = 'Wavestrider'
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        """
        alway invalid to directly use.
        """
        return False

    def get_actions(self, match: Any) -> List[MakeDamageAction]:
        return [
            self.attack_opposite_active(match, self.damage, self.damage_type)
        ]


class Stormbreaker(ElementalBurstBase):
    name: Literal['Stormbreaker'] = 'Stormbreaker'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
        charge = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_team_status("Thunderbeast's Targe"),
        ]


# Talents


class LightningStorm_4_2(SkillTalent):
    name: Literal['Lightning Storm']
    version: Literal['4.2'] = '4.2'
    charactor_name: Literal['Beidou'] = 'Beidou'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
    )
    skill: Literal['Tidecaller'] = 'Tidecaller'

    usage: int = 2
    max_usage: int = 2
    activated: bool = False
    need_to_activate: bool = False

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        reset usage and activate
        """
        self.usage = self.max_usage
        self.activated = False
        return []

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        If self receive damage and has status, activate
        """
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped
            return []
        if not event.final_damage.is_corresponding_charactor_receive_damage(
            self.position, match,
        ):
            # not corresponding charactor
            return []
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        for status in charactor.status:
            if status.name == 'Tidecaller: Surf Embrace':
                self.activated = True
                break
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL'],
    ) -> CostValue:
        """
        If activated, decrease self normal attack cost by 1
        """
        if not self.activated and self.need_to_activate:
            # not activated and need to activate
            return value
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
            source_area = ObjectPositionType.CHARACTOR,
        ):
            # not self skill
            return value
        if value.cost.label & CostLabels.NORMAL_ATTACK.value == 0:
            # not normal attack
            return value
        if self.usage <= 0:
            # no usage
            return value
        # decrease
        if value.cost.decrease_cost(None):  # pragma: no branch
            if mode == 'REAL':
                self.usage -= 1
        return value


# charactor base


class Beidou_3_8(CharactorBase):
    name: Literal['Beidou']
    version: Literal['3.8'] = '3.8'
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | Tidecaller | Wavestrider | Stormbreaker
    ] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Oceanborne',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            Tidecaller(),
            Wavestrider(),
            Stormbreaker(),
        ]


register_class(Beidou_3_8 | LightningStorm_4_2)



from typing import Any, List, Literal

from ....utils.class_registry import register_base_class, register_class
from ...dice import Dice

from ...action import (
    Actions, CreateDiceAction, DrawCardAction, 
    GenerateRerollDiceRequestAction, MakeDamageAction, RemoveDiceAction, 
    RemoveObjectAction
)
from ...event import (
    MoveObjectEventArguments, PlayerActionStartEventArguments, 
    SkillEndEventArguments, UseCardEventArguments
)
from ...modifiable_values import (
    CostValue, DamageValue, InitialDiceColorValue, RerollValue
)
from ...event import RoundEndEventArguments, RoundPrepareEventArguments

from ...struct import Cost
from ...consts import (
    ELEMENT_DEFAULT_ORDER, ELEMENT_TO_DIE_COLOR, CostLabels, 
    DamageElementalType, DamageType, DieColor, IconType, ObjectPositionType
)
from .base import (
    RoundEffectSupportBase, SupportBase, UsageWithRoundRestrictionSupportBase
)


class LocationBase(SupportBase):
    cost_label: int = CostLabels.CARD.value | CostLabels.LOCATION.value


register_base_class(LocationBase)


class RoundEffectLocationBase(RoundEffectSupportBase, LocationBase):
    cost_label: int = CostLabels.CARD.value | CostLabels.LOCATION.value


class LiyueHarborWharf_3_3(LocationBase):
    name: Literal['Liyue Harbor Wharf']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 2)
    usage: int = 2
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[DrawCardAction | RemoveObjectAction]:
        """
        When in round end, draw 2 cards, and check if should remove.
        """
        if self.position.area != 'SUPPORT':
            # not in support area, do nothing
            return []
        self.usage -= 1
        return [
            DrawCardAction(
                player_idx = self.position.player_idx,
                number = 2,
                draw_if_filtered_not_enough = True
            ),
        ] + self.check_should_remove()


class KnightsOfFavoniusLibrary_3_3(LocationBase):
    name: Literal['Knights of Favonius Library']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 1)
    usage: int = 0
    icon_type: Literal[IconType.NONE] = IconType.NONE

    def play(self, match: Any) -> List[GenerateRerollDiceRequestAction]:
        return [GenerateRerollDiceRequestAction(
            player_idx = self.position.player_idx,
            reroll_times = 1,
        )]

    def value_modifier_REROLL(
        self, value: RerollValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> RerollValue:
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return value
        if value.player_idx != self.position.player_idx:
            # not self player
            return value
        value.value += 1
        return value


class JadeChamber_4_0(LocationBase):
    name: Literal['Jade Chamber']
    version: Literal['4.0'] = '4.0'
    cost: Cost = Cost()
    usage: int = 0
    icon_type: Literal[IconType.NONE] = IconType.NONE

    def value_modifier_INITIAL_DICE_COLOR(
            self, value: InitialDiceColorValue, 
            match: Any,
            mode: Literal['REAL', 'TEST']) -> InitialDiceColorValue:
        """
        If self in support, fix two dice colors to self
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self player
            return value
        active = match.player_tables[
            self.position.player_idx].get_active_charactor()
        element = active.element
        value.dice_colors += [ELEMENT_TO_DIE_COLOR[element]] * 2
        return value


class JadeChamber_3_3(JadeChamber_4_0):
    version: Literal['3.3']
    cost: Cost = Cost(same_dice_number = 1)


class DawnWinery_3_3(RoundEffectLocationBase):
    name: Literal['Dawn Winery']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 2)
    max_usage_per_round: int = 1

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self player
            return value
        if self.usage <= 0:
            # no usage
            return value
        if value.cost.label & CostLabels.SWITCH_CHARACTOR.value == 0:
            # not switch character
            return value
        # decrease cost
        if value.cost.decrease_cost(None):
            if mode == 'REAL':
                self.usage -= 1
        return value


class WangshuInn_3_3(LocationBase):
    name: Literal['Wangshu Inn']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 2)
    usage: int = 2
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        If back charactor not full hp, heal 2 hp, and check if should remove.
        """
        if self.position.area != 'SUPPORT':
            # not in support area, do nothing
            return []
        table = match.player_tables[self.position.player_idx]
        charactors = table.charactors
        active = table.active_charactor_idx
        charactor = None
        for cnum, c in enumerate(charactors):
            if cnum == active or c.is_defeated:
                # active or defeated, do nothing
                continue
            if charactor is None or c.damage_taken > charactor.damage_taken:
                charactor = c
        if charactor is None:
            # no charactor on standby, do nothing
            return []
        if charactor.damage_taken <= 0:
            # full hp, do nothing
            return []
        self.usage -= 1
        return [MakeDamageAction(
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = charactor.position,
                    damage = -2,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = self.cost.copy()
                )
            ],
        )] + self.check_should_remove()


class FavoniusCathedral_3_3(LocationBase):
    name: Literal['Favonius Cathedral']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 2)
    usage: int = 2
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        If active charactor not full hp, heal 2 hp, and check if should remove.
        """
        if self.position.area != 'SUPPORT':
            # not in support area, do nothing
            return []
        charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        if charactor.damage_taken <= 0:
            # full hp, do nothing
            return []
        self.usage -= 1
        return [MakeDamageAction(
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = charactor.position,
                    damage = -2,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = self.cost.copy()
                )
            ],
        )] + self.check_should_remove()


class Tenshukaku_3_7(LocationBase):
    name: Literal['Tenshukaku']
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 2)
    usage: int = 0
    icon_type: Literal[IconType.NONE] = IconType.NONE

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        When in round prepare, if have 5 different kinds of elemental dice,
        create 1 omni element.
        """
        if self.position.area != 'SUPPORT':
            # not in support area, do nothing
            return []
        colors = match.player_tables[self.position.player_idx].dice.colors
        omni_number = 0
        color_set = set()
        for color in colors:
            if color == DieColor.OMNI:
                omni_number += 1
            else:
                color_set.add(color)
        if len(color_set) + omni_number < 5:
            # not enough different kinds of elemental dice
            return []
        return [CreateDiceAction(
            player_idx = self.position.player_idx,
            number = 1,
            color = DieColor.OMNI
        )]


class GrandNarukamiShrine_3_6(LocationBase):
    name: Literal['Grand Narukami Shrine']
    version: Literal['3.6'] = '3.6'
    cost: Cost = Cost(same_dice_number = 2)
    usage: int = 3
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def play(self, match: Any) -> List[Actions]:
        """
        When played, set usage to 2 and generate one random die.
        """
        self.usage = 2
        return [CreateDiceAction(
            player_idx = self.position.player_idx,
            number = 1,
            different = True  # use different to avoid OMNI die
        )]

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateDiceAction | RemoveObjectAction]:
        """
        create one random die, and out of usage remove self.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        self.usage -= 1
        return [CreateDiceAction(
            player_idx = self.position.player_idx,
            number = 1,
            different = True  # use different to avoid OMNI die
        )] + self.check_should_remove()


class SangonomiyaShrine_3_7(LocationBase):
    name: Literal['Sangonomiya Shrine']
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 2)
    usage: int = 2
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        If active charactor not full hp, heal 2 hp, and check if should remove.
        """
        if self.position.area != 'SUPPORT':
            # not in support area, do nothing
            return []
        charactors = match.player_tables[self.position.player_idx].charactors
        damage_action = MakeDamageAction(
            damage_value_list = [],
        )
        for charactor in charactors:
            if charactor.is_defeated or charactor.damage_taken <= 0:
                # full hp or defeated, do nothing
                continue
            damage_action.damage_value_list.append(
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = charactor.position,
                    damage = -1,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = self.cost.copy()
                )
            )
        if len(damage_action.damage_value_list) == 0:
            # no charactor to heal, do nothing
            return []
        self.usage -= 1
        return [damage_action] + self.check_should_remove()


class SumeruCity_3_7(RoundEffectLocationBase):
    name: Literal['Sumeru City']
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 2)
    max_usage_per_round: int = 1

    def play(self, match: Any) -> List[Actions]:
        """
        When played, reset usage.
        """
        self.usage = 1
        return super().play(match)

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        When self charactor use skills or equip talents, and have usage,
        and have less or equal elemental dice than cards in hand, reduce cost.
        """
        if self.position.area != 'SUPPORT':
            # not in support area, do nothing
            return value
        if self.usage == 0:
            # no usage
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self player
            return value
        label = (
            CostLabels.NORMAL_ATTACK.value | CostLabels.ELEMENTAL_SKILL.value
            | CostLabels.ELEMENTAL_BURST.value | CostLabels.TALENT.value
        )
        if value.cost.label & label == 0:
            # not skill or talent
            return value
        table = match.player_tables[self.position.player_idx]
        if len(table.dice.colors) > len(table.hands):
            # more elemental dice than cards in hand
            return value
        # reduce cost
        if value.cost.decrease_cost(value.cost.elemental_dice_color):
            # success
            if mode == 'REAL':
                self.usage -= 1
        return value


class Vanarana_3_7(LocationBase):
    name: Literal['Vanarana']
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost()
    usage: int = 0
    colors: List[DieColor] = []
    icon_type: Literal[IconType.COUNTER] = IconType.COUNTER

    def play(self, match: Any) -> List[Actions]:
        self.usage = 0
        self.colors = []
        return super().play(match)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        When in round prepare, give collected dice back.
        """
        assert self.usage == len(self.colors)
        if self.position.area != 'SUPPORT':
            # not in support area, do nothing
            return []
        if self.usage == 0:
            # no dice, do nothing
            return []
        ret: List[CreateDiceAction] = []
        for color in self.colors:
            ret.append(CreateDiceAction(
                player_idx = self.position.player_idx,
                color = color,
                number = 1
            ))
        self.usage = 0
        self.colors = []
        return ret

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[RemoveDiceAction]:
        """
        When in round end, collect up to 2 unused elemental dice.
        """
        if self.position.area != 'SUPPORT':
            # not in support area, do nothing
            return []
        assert self.usage == 0 and len(self.colors) == 0
        current_dice: Dice = match.player_tables[
            self.position.player_idx].dice
        current_dice_colors = current_dice.colors
        if len(current_dice_colors) == 0:
            # no dice, do nothing
            return []
        dice_map = {}
        for color in current_dice_colors:
            if color not in dice_map:
                dice_map[color] = 0
            dice_map[color] += 1
        # first try to gather two same element dice
        for element in ELEMENT_DEFAULT_ORDER:
            color = ELEMENT_TO_DIE_COLOR[element]
            if (
                color in dice_map 
                and dice_map[color] >= 2
                and len(self.colors) < 2
            ):
                self.colors = [color, color]
        if len(self.colors) < 2:
            # if not enough, gather any two dice
            for element in ELEMENT_DEFAULT_ORDER:
                color = ELEMENT_TO_DIE_COLOR[element]
                if color in dice_map and len(self.colors) < 2:
                    self.colors.append(color)
        if len(self.colors) < 2:
            # if not enough, gather omni
            if DieColor.OMNI in dice_map:
                self.colors.append(DieColor.OMNI)
                if len(self.colors) < 2 and dice_map[DieColor.OMNI] >= 2:
                    self.colors.append(DieColor.OMNI)
        self.usage = len(self.colors)
        dice_idxs = current_dice.colors_to_idx(self.colors)
        return [RemoveDiceAction(
            player_idx = self.position.player_idx,
            dice_idxs = dice_idxs,
        )]


class ChinjuForest_3_7(LocationBase):
    name: Literal['Chinju Forest']
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 1)
    usage: int = 3
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateDiceAction | RemoveObjectAction]:
        """
        When in round prepare, if have usage, create one die that match
        active charactor element.
        """
        if self.position.area != ObjectPositionType.SUPPORT:
            # not in support area, do nothing
            return []
        if self.usage <= 0:  # pragma: no cover
            # no usage
            return []
        if event.player_go_first == self.position.player_idx:
            # self go first, do nothing
            return []
        table = match.player_tables[self.position.player_idx]
        active = table.get_active_charactor()
        element = active.element
        self.usage -= 1
        return [CreateDiceAction(
            player_idx = self.position.player_idx,
            number = 1,
            color = ELEMENT_TO_DIE_COLOR[element]
        )] + self.check_should_remove()


class GoldenHouse_4_0(LocationBase, UsageWithRoundRestrictionSupportBase):
    name: Literal['Golden House']
    version: Literal['4.0'] = '4.0'
    cost: Cost = Cost()
    usage: int = 2
    max_usage_one_round: int = 1
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    card_cost_label: int = CostLabels.WEAPON.value
    decrease_threshold: int = 3

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        if card label match and original cost greater than threshold, 
        reduce cost.
        """
        if (
            self.position.area == ObjectPositionType.SUPPORT
            and value.position.player_idx == self.position.player_idx
            and value.cost.original_value.total_dice_cost
            >= self.decrease_threshold
            and value.cost.label & self.card_cost_label != 0
            and self.has_usage()
        ):
            # area right, player right, cost label match, cost greater than
            # threshold, and not used this round
            if value.cost.decrease_cost(None):
                if mode == 'REAL':
                    self.use()
        return value

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        if self.usage == 0:
            # if usage is zero, it run out of usage, remove it
            return list(self.check_should_remove())
        # otherwise, do parent event handler
        return super().event_handler_MOVE_OBJECT(event, match)


class GandharvaVille_4_1(LocationBase, UsageWithRoundRestrictionSupportBase):
    name: Literal['Gandharva Ville']
    version: Literal['4.1'] = '4.1'
    cost: Cost = Cost(same_dice_number = 1)
    usage: int = 3
    max_usage_one_round: int = 1
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_PLAYER_ACTION_START(
        self, event: PlayerActionStartEventArguments, match: Any
    ) -> List[Actions]:
        """
        If self action start, and has usage, and no elemental dice, create
        one omni element.
        """
        if not (
            self.position.area == ObjectPositionType.SUPPORT
            and event.player_idx == self.position.player_idx
            and self.has_usage()
            and len(match.player_tables[
                self.position.player_idx].dice.colors) == 0
        ):
            # not equipped, or not self, or no usage, or have elemental dice
            return []
        # create one omni element
        self.use()
        return [CreateDiceAction(
            player_idx = self.position.player_idx,
            number = 1,
            color = DieColor.OMNI
        )]


class StormterrorsLair_4_2(LocationBase, UsageWithRoundRestrictionSupportBase):
    name: Literal["Stormterror's Lair"]
    version: Literal['4.2'] = '4.2'
    cost: Cost = Cost(same_dice_number = 2)
    usage: int = 3
    max_usage_one_round: int = 1
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    card_cost_label: int = CostLabels.SKILLS.value
    decrease_threshold: int = 4

    def play(self, match: Any) -> List[Actions]:
        """
        Draw a talent card
        """
        return super().play(match) + [DrawCardAction(
            player_idx = self.position.player_idx,
            number = 1,
            whitelist_cost_labels = CostLabels.TALENT.value,
            draw_if_filtered_not_enough = False,
        )]

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        if card label match and original cost greater than threshold, 
        or is talent cost, reduce cost.
        """
        if (
            self.position.area == ObjectPositionType.SUPPORT
            and value.position.player_idx == self.position.player_idx
            and self.has_usage()
        ):
            # area right, player right, and not used this round
            if (
                (
                    value.cost.original_value.total_dice_cost
                    >= self.decrease_threshold
                    and value.cost.label & self.card_cost_label != 0
                )
                or value.cost.label & CostLabels.TALENT.value != 0
            ):
                # cost label match, cost greater than threshold; or is talent
                if value.cost.decrease_cost(value.cost.elemental_dice_color):
                    if mode == 'REAL':
                        self.use()
        return value

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Any
    ) -> List[Actions]:
        if self.usage == 0:
            # if usage is zero, it run out of usage, remove it
            return list(self.check_should_remove())
        # otherwise, do parent event handler
        return super().event_handler_USE_CARD(event, match)

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        if self.usage == 0:
            # if usage is zero, it run out of usage, remove it
            return list(self.check_should_remove())
        return []


register_class(
    LiyueHarborWharf_3_3 | KnightsOfFavoniusLibrary_3_3 | JadeChamber_4_0 
    | JadeChamber_3_3 | DawnWinery_3_3 | WangshuInn_3_3 
    | FavoniusCathedral_3_3 | Tenshukaku_3_7 | GrandNarukamiShrine_3_6 
    | SangonomiyaShrine_3_7 | SumeruCity_3_7 | Vanarana_3_7 | ChinjuForest_3_7 
    | GoldenHouse_4_0 | GandharvaVille_4_1 | StormterrorsLair_4_2
)



from typing import List, Literal

from ....utils.class_registry import register_class

from ..charactor_base import PhysicalNormalAttackBase
from ...struct import Cost
from ...consts import DieColor
from .kamisato_ayato_4_1 import KamisatoAyato_4_1 as KA_4_1
from .kamisato_ayato_4_1 import KamisatoArtSuiyuu as KAS_4_1
from .kamisato_ayato_4_1 import KamisatoArtKyouka


class KamisatoArtSuiyuu(KAS_4_1):
    damage: int = 3
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3,
        charge = 3
    )


class KamisatoAyato_3_6(KA_4_1):
    version: Literal['3.6']
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | KamisatoArtKyouka | KamisatoArtSuiyuu
    ] = []

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Kamisato Art: Marobashi',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            KamisatoArtKyouka(),
            KamisatoArtSuiyuu(),
        ]


register_class(KamisatoAyato_3_6)



from src.lpsim.agents import InteractionAgent
from src.lpsim import Deck, Match, MatchState
from tests.utils_for_test import (
    get_random_state, get_test_id_from_command, make_respond, 
    set_16_omni
)


def test_fishchip_gilded():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 1 15 14 13",
            "sw_char 1 12",
            "TEST 1 p0 dice 12",
            "card 1 1 11 10 9",
            "TEST 2 p0 hand 4",
            "TEST 1 p0 dice 11",
            "skill 1 10 9 8",
            "sw_char 2 7",
            "card 2 2 6 5 4",
            "skill 0 5 4 3",
            "skill 0 2 1 0",
            "end",
            "TEST 2 p0 hand 7",
            "skill 0 15 14 13",
            "sw_char 1 12",
            "card 4 1",
            "card 4 0",
            "skill 1 11 10 9",
            "end"
        ],
        [
            "sw_card 2",
            "choose 0",
            "card 2 0",
            "card 1 0 15 14",
            "skill 1 13 12 11",
            "sw_char 1 10",
            "skill 1 9 8",
            "sw_char 0 7",
            "TEST 2 p0 hand 4",
            "skill 0 6 5 4",
            "TEST 2 p0 hand 5",
            "sw_char 2 3",
            "skill 0 2 1",
            "end",
            "TEST 2 p0 hand 8",
            "sw_char 1 15",
            "card 2 2 14 13 12",
            "TEST 1 p1 dice 13",
            "TEST 3 p1 omni*12 dendro*1",
            "end",
            "TEST 2 p0 hand 9",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck_str_1 = '''
        default_version:4.3
        charactor:Fischl
        charactor:Mona
        charactor:Nahida
        Fish and Chips*10
        Sweet Madame*10
        Gilded Dreams*10
    '''
    deck_str_2 = '''
        default_version:4.3
        charactor:Barbara
        charactor:Mona
        charactor:Nahida
        Fish and Chips*10
        Sweet Madame*10
        Gilded Dreams*10
    '''
    match.set_deck([Deck.from_str(deck_str_1), Deck.from_str(deck_str_2)])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(' ')
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[
                    pidx].dice.colors) == int(cmd[4])
            elif test_id == 2:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].hands) == int(cmd[4])
            elif test_id == 3:
                pidx = int(cmd[2][1])
                colors = {}
                for c in cmd[3:]:
                    c, n = c.strip().split('*')
                    colors[c] = int(n)
                for c in match.player_tables[pidx].dice.colors:
                    c = c.value.lower()
                    assert c in colors
                    colors[c] -= 1
                    if colors[c] == 0:
                        del colors[c]
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == '__main__':
    test_fishchip_gilded()



import os
from src.lpsim import MatchState
from tests.utils_for_test import (
    check_hp, check_usage, get_test_id_from_command, make_respond, 
    read_from_log_json
)


def test_lynette():
    match, agent_0, agent_1 = read_from_log_json(
        os.path.join(os.path.dirname(__file__), 'jsons', 'test_lynette.json')
    )
    match.start()
    match.step()
    new_commands = [[], []]
    while True:
        if match.need_respond(0):
            agent = agent_0
            nc = new_commands[0]
        elif match.need_respond(1):
            agent = agent_1
            nc = new_commands[1]
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            nc.append(agent.commands[0])
            cmd = agent.commands[0].strip().split(' ')
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:6], hps[6:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].summons, cmd[3:])
            elif test_id == 4:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].team_status, cmd[3:])
            elif test_id == 9:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].summons) == len(cmd[3:])
                for i, s in enumerate(match.player_tables[pidx].summons):
                    assert s.damage_elemental_type.value == cmd[3 + i]
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == '__main__':
    test_lynette()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...event import RoundEndEventArguments

from ...action import (
    Actions, ChangeObjectUsageAction, ChargeAction, MakeDamageAction
)
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class AfterSalesServiceRounds_4_2(AttackerSummonBase):
    name: Literal['After-Sales Service Rounds'] = 'After-Sales Service Rounds'
    version: Literal['4.2'] = '4.2'
    usage: int = 1
    max_usage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1


class Jinni_4_2(AttackerSummonBase):
    name: Literal['Jinni'] = 'Jinni'
    version: Literal['4.2'] = '4.2'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.HEAL
    damage: int = -2
    talent_activated: bool = False

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChargeAction | ChangeObjectUsageAction]:
        active_charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        charge = 1
        if self.talent_activated:
            if active_charactor.hp <= 6:
                self.damage = -3
            if active_charactor.charge == 0:
                charge = 2
        ret = super().event_handler_ROUND_END(event, match) + [
            ChargeAction(
                player_idx = self.position.player_idx,
                charactor_idx = active_charactor.position.charactor_idx,
                charge = charge,
            )
        ]
        self.damage = -2
        return ret


# Skills


class SpiritWaridingLampTroubleshooterCannon(ElementalSkillBase):
    name: Literal[
        'Spirit-Warding Lamp: Troubleshooter Cannon'
    ] = 'Spirit-Warding Lamp: Troubleshooter Cannon'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_summon('After-Sales Service Rounds')]


class AlcazarzaraysExactitude(ElementalBurstBase):
    name: Literal["Alcazarzaray's Exactitude"] = "Alcazarzaray's Exactitude"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_summon('Jinni', {
                'talent_activated': self.is_talent_equipped(match)
            })]


# Talents


class DiscretionarySupplement_4_2(SkillTalent):
    name: Literal['Discretionary Supplement']
    version: Literal['4.2'] = '4.2'
    charactor_name: Literal['Dori'] = 'Dori'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
        charge = 2
    )
    skill: Literal["Alcazarzaray's Exactitude"] = "Alcazarzaray's Exactitude"


# charactor base


class Dori_4_2(CharactorBase):
    name: Literal['Dori']
    version: Literal['4.2'] = '4.2'
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | SpiritWaridingLampTroubleshooterCannon 
        | AlcazarzaraysExactitude
    ] = []
    faction: List[FactionType] = [
        FactionType.SUMERU
    ]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Marvelous Sword-Dance (Modified)',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SpiritWaridingLampTroubleshooterCannon(),
            AlcazarzaraysExactitude()
        ]


register_class(
    Dori_4_2 | DiscretionarySupplement_4_2 | Jinni_4_2 
    | AfterSalesServiceRounds_4_2
)



from src.lpsim.agents.interaction_agent import InteractionAgent
from src.lpsim.server.match import Match, MatchState
from src.lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp, get_random_state, get_test_id_from_command, make_respond, 
    set_16_omni
)


def test_yoimiya_3_3():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 1 0",
            "skill 0 0 1 2",
            "skill 1 0",
            "card 0 0 0 1",
            "skill 1 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 0 0 1 2",
            "end",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "TEST 1 8 10 10 3 6 7",
            "TEST 2 p1 all pyro app",
            "skill 2 0 1 2",
            "TEST 1 8 10 10 0 6 7",
            "skill 0 0 1 2",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "TEST 1 8 10 10 0 6 2",
            "TEST 3 p0 status usage 2",
            "end",
            "skill 1 0 1 2"
        ],
        [
            "sw_card",
            "choose 2",
            "sw_char 1 0",
            "sw_char 2 0",
            "TEST 1 10 10 10 10 10 7",
            "sw_char 1 0",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "sw_char 1 0",
            "TEST 1 8 10 10 10 6 7",
            "sw_char 0 0",
            "end",
            "end",
            "choose 2",
            "skill 1 0 1 2",
            "choose 1",
            "TEST 1 8 9 10 0 5 0",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.0
        charactor:Yoimiya@3.3
        charactor:Nahida@3.7
        charactor:Mona@3.3
        Naganohara Meteor Swarm*30
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(' ')[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                for charactor in match.player_tables[1].charactors:
                    assert charactor.element_application == ['PYRO']
            elif test_id == 3:
                assert len(match.player_tables[0].team_status) == 1
                assert match.player_tables[0].team_status[0].usage == 2
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_yoimiya_different_version_generate():
    cmd_records = [
        [
            "TEST 1 3.3 3.4 3.8 Yoimiya",
            "sw_card"
        ],
        [
            "sw_card"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.0
        charactor:Yoimiya@3.3
        charactor:Yoimiya@3.4
        charactor:Yoimiya
        Send Off@3.3*15
        Send Off*15
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                charactors = match.player_tables[0].charactors
                assert charactors[0].version == '3.3'
                assert charactors[1].version == '3.4'
                assert charactors[2].version == '3.8'
                assert charactors[0].skills[2].cost.charge == 2
                assert charactors[1].skills[2].cost.charge == 3
                assert charactors[2].skills[2].cost.charge == 3
                assert charactors[0].skills[2].cost.elemental_dice_number == 3
                assert charactors[1].skills[2].cost.elemental_dice_number == 4
                assert charactors[2].skills[2].cost.elemental_dice_number == 3
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == '__main__':
    test_yoimiya_3_3()
    test_yoimiya_different_version_generate()




from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....event import SkillEndEventArguments

from ....modifiable_values import DamageIncreaseValue

from ....action import CreateObjectAction

from ....consts import ObjectPositionType, SkillType, WeaponType

from ....struct import Cost, ObjectPosition
from .base import RoundEffectWeaponBase, WeaponBase


class AmosBow_3_7(RoundEffectWeaponBase):
    name: Literal["Amos' Bow"]
    cost: Cost = Cost(same_dice_number = 3)
    version: Literal['3.7'] = '3.7'
    weapon_type: WeaponType = WeaponType.BOW
    max_usage_per_round: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, 
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        condition_satisfied = False
        if (
            value.cost.total_dice_cost + value.cost.charge >= 5
            and self.usage > 0
        ):
            # have usage and costs satisfied
            self.damage_increase = 3
            condition_satisfied = True
        current_damage = value.damage
        super().value_modifier_DAMAGE_INCREASE(value, match, mode)
        if value.damage > current_damage:
            # value modified
            if condition_satisfied:
                # decrease usage
                assert mode == 'REAL'
                self.usage -= 1
        self.damage_increase = 1
        return value


class ElegyForTheEnd_3_7(WeaponBase):
    name: Literal["Elegy for the End"]
    cost: Cost = Cost(same_dice_number = 3)
    version: Literal['3.7'] = '3.7'
    weapon_type: WeaponType = WeaponType.BOW

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If equipped and self use elemental burst, create a status
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, 
            source_area = ObjectPositionType.CHARACTOR
        ):
            # not equipped or not self use skill
            return []
        if event.action.skill_type != SkillType.ELEMENTAL_BURST:
            # not elemental burst
            return []
        return [CreateObjectAction(
            object_name = 'Millennial Movement: Farewell Song',
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectPositionType.TEAM_STATUS,
                id = 0
            ),
            object_arguments = {}
        )]


register_class(AmosBow_3_7 | ElegyForTheEnd_3_7)



"""
Event cards that not belong to any other categories.
"""

from typing import Any, List, Literal, Tuple

from ....utils.class_registry import register_class

from ...event import (
    CharactorDefeatedEventArguments, RoundPrepareEventArguments
)

from ...consts import (
    DieColor, ObjectPositionType, ObjectType, PlayerActionLabels, SkillType
)

from ...object_base import EventCardBase, MultiTargetEventCardBase
from ...action import (
    ActionTypes, Actions, ChangeObjectUsageAction, ChargeAction, 
    CreateDiceAction, CreateObjectAction, DrawCardAction, 
    GenerateRerollDiceRequestAction, MoveObjectAction, RemoveObjectAction, 
    SkillEndAction, SwitchCharactorAction, UseSkillAction
)
from ...struct import Cost, MultipleObjectPosition, ObjectPosition


class TheBestestTravelCompanion_3_3(EventCardBase):
    name: Literal['The Bestest Travel Companion!']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(
        any_dice_number = 2
    )

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateDiceAction]:
        """
        Act the card. Convert the Elemental Dice spent to Omni Element x2.
        """
        assert target is None  # no targets
        return [CreateDiceAction(
            player_idx = self.position.player_idx,
            color = DieColor.OMNI,
            number = 2,
        )]


class ChangingShifts_3_3(EventCardBase):
    name: Literal['Changing Shifts']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        """
        When only one charactor is alive, cannot use this card.
        """
        charactors = match.player_tables[self.position.player_idx].charactors
        counter = 0
        for charactor in charactors:
            if charactor.is_alive:
                counter += 1
        return counter > 1

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Act the card. Create team status.
        """
        assert target is None  # no targets
        return [CreateObjectAction(
            object_name = self.name,
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectPositionType.TEAM_STATUS,
                id = -1,
            ),
            object_arguments = {}
        )]


class TossUp_3_3(EventCardBase):
    name: Literal['Toss-Up']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost()

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[GenerateRerollDiceRequestAction]:
        assert target is None
        return [GenerateRerollDiceRequestAction(
            player_idx = self.position.player_idx,
            reroll_times = 2,
        )]


class Strategize_3_3(EventCardBase):
    name: Literal['Strategize']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(
        same_dice_number = 1
    )

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[DrawCardAction]:
        """
        Act the card. Draw two cards.
        """
        assert target is None  # no targets
        return [DrawCardAction(
            player_idx = self.position.player_idx, 
            number = 2,
            draw_if_filtered_not_enough = True
        )]


class IHaventLostYet_4_0(EventCardBase):
    name: Literal["I Haven't Lost Yet!"]
    version: Literal['4.0'] = '4.0'
    cost: Cost = Cost()

    activated: bool = False

    available_handler_in_deck: List[ActionTypes] = [
        ActionTypes.CHARACTOR_DEFEATED, 
        ActionTypes.ROUND_PREPARE
    ]

    def is_valid(self, match: Any) -> bool:
        team_status = match.player_tables[self.position.player_idx].team_status
        for status in team_status:
            if status.name == self.name:
                # activated in this round
                return False
        return self.activated

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ChargeAction | CreateDiceAction | CreateObjectAction]:
        """
        1 omni die, 1 energy, and generate used status.
        """
        assert target is None
        return [
            CreateDiceAction(
                player_idx = self.position.player_idx,
                color = DieColor.OMNI,
                number = 1,
            ),
            ChargeAction(
                player_idx = self.position.player_idx,
                charactor_idx = match.player_tables[
                    self.position.player_idx].active_charactor_idx,
                charge = 1
            ),
            CreateObjectAction(
                object_name = self.name,
                object_position = ObjectPosition(
                    player_idx = self.position.player_idx,
                    area = ObjectPositionType.TEAM_STATUS,
                    id = -1,
                ),
                object_arguments = {}
            )
        ]

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        Reset activated.
        """
        self.activated = False
        return []

    def event_handler_CHARACTOR_DEFEATED(
        self, event: CharactorDefeatedEventArguments, match: Any
    ) -> List[Actions]:
        """
        Mark activated.
        """
        if self.position.player_idx != event.action.player_idx:
            return []
        self.activated = True
        return []


class IHaventLostYet_3_3(IHaventLostYet_4_0):
    name: Literal["I Haven't Lost Yet!"]
    version: Literal['3.3']
    cost: Cost = Cost()

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ChargeAction | CreateDiceAction | CreateObjectAction]:
        ret = super().get_actions(target, match)
        assert ret[-1].type == ActionTypes.CREATE_OBJECT
        # old version not adding team status
        ret = ret[:-1]
        return ret


class LeaveItToMe_3_3(EventCardBase):
    name: Literal['Leave It to Me!']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        """
        When only one charactor is alive, cannot use this card.
        """
        charactors = match.player_tables[self.position.player_idx].charactors
        counter = 0
        for charactor in charactors:
            if charactor.is_alive:
                counter += 1
        return counter > 1

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Act the card. Create team status.
        """
        assert target is None  # no targets
        return [CreateObjectAction(
            object_name = self.name,
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectPositionType.TEAM_STATUS,
                id = -1,
            ),
            object_arguments = {}
        )]


class WhenTheCraneReturned_3_3(EventCardBase):
    name: Literal['When the Crane Returned']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 1)

    def is_valid(self, match: Any) -> bool:
        """
        When only one charactor is alive, cannot use this card.
        """
        charactors = match.player_tables[self.position.player_idx].charactors
        counter = 0
        for charactor in charactors:
            if charactor.is_alive:
                counter += 1
        return counter > 1

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Act the card. Create team status.
        """
        assert target is None  # no targets
        return [CreateObjectAction(
            object_name = self.name,
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectPositionType.TEAM_STATUS,
                id = -1,
            ),
            object_arguments = {}
        )]


class Starsigns_3_3(EventCardBase):
    name: Literal['Starsigns']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(any_dice_number = 2)

    def is_valid(self, match: Any) -> bool:
        """
        can use if charge not full
        """
        active_charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        return active_charactor.charge < active_charactor.max_charge

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # active charactor
        return [match.player_tables[
            self.position.player_idx].get_active_charactor().position]

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ChargeAction]:
        assert target is not None
        return [ChargeAction(
            player_idx = target.player_idx,
            charactor_idx = target.charactor_idx,
            charge = 1,
        )]


class ClaxsArts_3_3(EventCardBase):
    name: Literal["Calx's Arts"]
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 1)

    def _get_charge_source_and_targets(
            self, match: Any) -> Tuple[int, List[int]]:
        """
        Get charge source and targets.
        """
        table = match.player_tables[self.position.player_idx]
        source = table.active_charactor_idx
        assert source >= 0 and source < len(table.charactors)
        targets = []
        for idx, charactor in enumerate(table.charactors):
            if idx != source and charactor.charge > 0:
                targets.append(idx)
        if len(targets) > 2:
            targets = targets[:2]
        return source, targets

    def is_valid(self, match: Any) -> bool:
        source, targets = self._get_charge_source_and_targets(match)
        return len(targets) > 0

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ChargeAction]:
        assert target is None
        source, targets = self._get_charge_source_and_targets(match)
        assert len(targets) > 0
        ret: List[ChargeAction] = []
        ret.append(ChargeAction(
            player_idx = self.position.player_idx,
            charactor_idx = source,
            charge = len(targets),
        ))
        for t in targets:
            ret.append(ChargeAction(
                player_idx = self.position.player_idx,
                charactor_idx = t,
                charge = -1,
            ))
        return ret


class MasterOfWeaponry_4_1(MultiTargetEventCardBase):
    name: Literal['Master of Weaponry']
    version: Literal['4.1'] = '4.1'
    cost: Cost = Cost()
    move_type: Literal[
        ObjectType.WEAPON, ObjectType.ARTIFACT
    ] = ObjectType.WEAPON
    reset_usage: bool = True

    def _move_target_id(self, charactor: Any) -> int | None:
        """
        If has target, return target id. otherwise return None
        """
        if self.move_type == ObjectType.WEAPON:
            return None if charactor.weapon is None else charactor.weapon.id
        elif self.move_type == ObjectType.ARTIFACT:
            return (
                None if charactor.artifact is None 
                else charactor.artifact.id
            )
        else:
            raise AssertionError('Unknown move type')

    def get_targets(self, match: Any) -> List[MultipleObjectPosition]:
        charactors = match.player_tables[self.position.player_idx].charactors
        ret: List[MultipleObjectPosition] = []
        for charactor in charactors:
            target_id = self._move_target_id(charactor)
            if target_id is not None:
                for target in charactors:
                    if target.id != charactor.id:
                        if (
                            self.move_type == ObjectType.WEAPON
                            and charactor.weapon_type != target.weapon_type
                        ):
                            # weapon type not right
                            continue
                        ret.append(MultipleObjectPosition(
                            positions = [
                                charactor.position.set_id(target_id),
                                target.position.set_id(target_id)
                            ]
                        ))
        return ret

    def is_valid(self, match: Any) -> bool:
        return len(self.get_targets(match)) > 0

    def get_actions(
        self, target: MultipleObjectPosition | None, match: Any
    ) -> List[RemoveObjectAction | MoveObjectAction]:
        assert target is not None
        assert len(target.positions) == 2
        ret: List[RemoveObjectAction | MoveObjectAction] = []
        target_charactor = match.player_tables[target.positions[
            1].player_idx].charactors[target.positions[1].charactor_idx]
        target_equip: Any = None
        if self.move_type == ObjectType.WEAPON:
            target_equip = target_charactor.weapon
        elif self.move_type == ObjectType.ARTIFACT:
            target_equip = target_charactor.artifact
        else:
            raise AssertionError('Unknown move type')
        if target_equip is not None:
            ret.append(RemoveObjectAction(
                object_position = target_equip.position
            ))
        return ret + [MoveObjectAction(
            object_position = target.positions[0],
            target_position = target.positions[1],
            reset_usage = self.reset_usage,
        )]


class BlessingOfTheDivineRelicsInstallation_4_1(MasterOfWeaponry_4_1):
    name: Literal["Blessing of the Divine Relic's Installation"]
    version: Literal['4.1'] = '4.1'
    cost: Cost = Cost()
    move_type: Literal[
        ObjectType.WEAPON, ObjectType.ARTIFACT
    ] = ObjectType.ARTIFACT
    reset_usage: bool = True


class MasterOfWeaponry_3_3(MasterOfWeaponry_4_1):
    version: Literal['3.3']
    reset_usage: bool = False


class BlessingOfTheDivineRelicsInstallation_3_3(
    BlessingOfTheDivineRelicsInstallation_4_1
):
    version: Literal['3.3']
    reset_usage: bool = False


class QuickKnit_3_3(EventCardBase):
    name: Literal['Quick Knit']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 1)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # all current summons
        return [x.position for x in match.player_tables[
            self.position.player_idx].summons]

    def is_valid(self, match: Any) -> bool:
        # should have summon
        return len(self.get_targets(match)) > 0

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        Add 1 usage of selected summon
        """
        assert target is not None
        return [ChangeObjectUsageAction(
            object_position = target,
            change_usage = 1
        )]


class SendOff_3_7(EventCardBase):
    name: Literal['Send Off']
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 2)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        Opposing summons
        """
        summons = match.player_tables[1 - self.position.player_idx].summons
        return [x.position for x in summons]

    def is_valid(self, match: Any) -> bool:
        return len(self.get_targets(match)) > 0

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        Act the card. Create team status.
        """
        assert target is not None
        return [ChangeObjectUsageAction(
            object_position = target,
            change_usage = -2
        )]


class SendOff_3_3(SendOff_3_7):
    name: Literal['Send Off']
    version: Literal['3.3']
    cost: Cost = Cost(any_dice_number = 2)

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[RemoveObjectAction]:
        """
        Act the card. Create team status.
        """
        assert target is not None
        return [RemoveObjectAction(
            object_position = target,
        )]


class GuardiansOath_3_3(EventCardBase):
    name: Literal["Guardian's Oath"]
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 4)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def is_valid(self, match: Any) -> bool:
        return (
            len(match.player_tables[0].summons) > 0
            or len(match.player_tables[1].summons) > 0
        )

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[RemoveObjectAction]:
        """
        Act the card. Destroy all summons.
        """
        assert target is None
        ret: List[RemoveObjectAction] = []
        for table in match.player_tables:
            for summon in table.summons:
                ret.append(RemoveObjectAction(
                    object_position = summon.position
                ))
        return ret


class PlungingStrike_3_7(EventCardBase):
    name: Literal['Plunging Strike']
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 3)

    def get_action_type(self, match: Any) -> Tuple[int, bool]:
        return (
            PlayerActionLabels.SWITCH.value
            | PlayerActionLabels.CARD.value
            | PlayerActionLabels.SKILL.value,
            True
        )

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        """
        pos: List[ObjectPosition] = []
        table = match.player_tables[self.position.player_idx]
        for cid, charactor in enumerate(table.charactors):
            if charactor.is_alive and cid != table.active_charactor_idx:
                pos.append(charactor.position)
        return pos

    def is_valid(self, match: Any) -> bool:
        return len(self.get_targets(match)) > 0

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[SwitchCharactorAction | UseSkillAction | SkillEndAction]:
        """
        Switch to target, and use target's normal attack
        """
        assert target is not None
        skills = match.player_tables[target.player_idx].charactors[
            target.charactor_idx].skills
        normal_idx: int = -1
        for idx, skill in enumerate(skills):
            if skill.skill_type == SkillType.NORMAL_ATTACK:
                assert normal_idx == -1, 'Multiple normal attack skill'
                normal_idx = idx
        return [
            SwitchCharactorAction(
                player_idx = target.player_idx,
                charactor_idx = target.charactor_idx
            ),
            UseSkillAction(
                skill_position = skills[normal_idx].position,
            ),
            SkillEndAction(
                position = skills[normal_idx].position,
                target_position = match.player_tables[
                    1 - target.player_idx].get_active_charactor().position,
                skill_type = SkillType.NORMAL_ATTACK
            )
        ]


class HeavyStrike_3_7(EventCardBase):
    name: Literal['Heavy Strike']
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 1)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # active charactor
        return [match.player_tables[
            self.position.player_idx].get_active_charactor().position]

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Act the card. Create charactor status.
        """
        assert target is not None
        return [CreateObjectAction(
            object_name = self.name,
            object_position = target.set_area(
                ObjectPositionType.CHARACTOR_STATUS),
            object_arguments = {}
        )]


class TheLegendOfVennessa_3_7(EventCardBase):
    name: Literal['The Legend of Vennessa']
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 3)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateDiceAction]:
        """
        Act the card. Create 4 basic Elemental Dice of different types.
        """
        assert target is None
        return [
            CreateDiceAction(
                player_idx = self.position.player_idx,
                number = 4,
                different = True
            ),
        ]


class FriendshipEternal_3_7(EventCardBase):
    name: Literal['Friendship Eternal']
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(
        same_dice_number = 2
    )

    def is_valid(self, match: Any) -> bool:
        # if both player have 4 or more cards, cannot use this card.
        for pidx, table in enumerate(match.player_tables):
            if len(table.hands) < 4:
                return True
            if len(table.hands) == 4 and pidx == self.position.player_idx:
                # for self, 4 card can use to draw one card
                return True
        return False

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[DrawCardAction]:
        """
        Act the card. Both player draw card until card number reach 4.
        """
        assert target is None  # no targets
        ret: List[DrawCardAction] = []
        for pidx, table in enumerate(match.player_tables):
            target_number = 4
            if len(table.hands) >= target_number:
                # already have enough cards, do nothing
                continue
            ret.append(DrawCardAction(
                player_idx = pidx,
                number = target_number - len(table.hands),
                draw_if_filtered_not_enough = True
            ))
        return ret


class RhythmOfTheGreatDream_3_8(EventCardBase):
    name: Literal['Rhythm of the Great Dream']
    version: Literal['3.8'] = '3.8'
    cost: Cost = Cost()

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        # no targets
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Act the card. Create team status.
        """
        assert target is None
        return [CreateObjectAction(
            object_name = self.name,
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectPositionType.TEAM_STATUS,
                id = -1,
            ),
            object_arguments = {}
        )]


class WhereIstheUnseenRazor_4_0(EventCardBase):
    name: Literal['Where Is the Unseen Razor?'] = 'Where Is the Unseen Razor?'
    version: Literal['4.0'] = '4.0'
    cost: Cost = Cost()

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        Self charactors that have weapon.
        """
        charactors = match.player_tables[self.position.player_idx].charactors
        res: List[ObjectPosition] = []
        for charactor in charactors:
            if charactor.weapon is not None:
                res.append(charactor.position)
        return res

    def is_valid(self, match: Any) -> bool:
        return len(self.get_targets(match)) > 0

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[MoveObjectAction | CreateObjectAction]:
        assert target is not None
        charactor = match.get_object(target)
        assert charactor is not None
        assert charactor.weapon is not None
        target_position = charactor.weapon.position.set_area(
            ObjectPositionType.HAND)
        return [
            MoveObjectAction(
                object_position = charactor.weapon.position,
                target_position = target_position
            ),
            CreateObjectAction(
                object_name = self.name,
                object_position = ObjectPosition(
                    player_idx = self.position.player_idx,
                    area = ObjectPositionType.TEAM_STATUS,
                    id = -1,
                ),
                object_arguments = {}
            )
        ]


class Pankration_4_1(EventCardBase):
    name: Literal['Pankration!']
    version: Literal['4.1'] = '4.1'
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        return (
            len(match.player_tables[self.position.player_idx].dice.colors) >= 8
            and not match.player_tables[
                1 - self.position.player_idx].has_round_ended
        )

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        create status in our side
        """
        assert target is None
        return [
            CreateObjectAction(
                object_position = ObjectPosition(
                    player_idx = self.position.player_idx,
                    area = ObjectPositionType.TEAM_STATUS,
                    id = -1,
                ),
                object_name = self.name,
                object_arguments = {}
            )
        ]


class Lyresong_4_2(EventCardBase):
    name: Literal['Lyresong'] = 'Lyresong'
    version: Literal['4.2'] = '4.2'
    cost: Cost = Cost()

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        Self charactors that have artifact.
        """
        charactors = match.player_tables[self.position.player_idx].charactors
        res: List[ObjectPosition] = []
        for charactor in charactors:
            if charactor.artifact is not None:
                res.append(charactor.position)
        return res

    def is_valid(self, match: Any) -> bool:
        return len(self.get_targets(match)) > 0

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[MoveObjectAction | CreateObjectAction]:
        assert target is not None
        charactor = match.get_object(target)
        assert charactor is not None
        assert charactor.artifact is not None
        target_position = charactor.artifact.position.set_area(
            ObjectPositionType.HAND)
        return [
            MoveObjectAction(
                object_position = charactor.artifact.position,
                target_position = target_position
            ),
            CreateObjectAction(
                object_name = self.name,
                object_position = ObjectPosition(
                    player_idx = self.position.player_idx,
                    area = ObjectPositionType.TEAM_STATUS,
                    id = -1,
                ),
                object_arguments = {}
            )
        ]


register_class(
    TheBestestTravelCompanion_3_3 | ChangingShifts_3_3 | TossUp_3_3 
    | Strategize_3_3 | IHaventLostYet_4_0 | LeaveItToMe_3_3 
    | WhenTheCraneReturned_3_3 | Starsigns_3_3 | ClaxsArts_3_3 
    | MasterOfWeaponry_4_1 | BlessingOfTheDivineRelicsInstallation_4_1 
    | QuickKnit_3_3 | SendOff_3_7 | GuardiansOath_3_3 | PlungingStrike_3_7 
    | HeavyStrike_3_7 | TheLegendOfVennessa_3_7 | FriendshipEternal_3_7 
    | RhythmOfTheGreatDream_3_8 | WhereIstheUnseenRazor_4_0 | Pankration_4_1 
    | Lyresong_4_2 | IHaventLostYet_3_3 | MasterOfWeaponry_3_3
    | BlessingOfTheDivineRelicsInstallation_3_3 | SendOff_3_3
)



from typing import Dict, List, Literal

from ....action import CreateObjectAction, SwitchCharactorAction

from .....utils.class_registry import register_class
from ....event import RoundEndEventArguments
from ....consts import IconType, ObjectPositionType
from ....status.charactor_status.base import CharactorStatusBase
from ....match import Match
from ....struct import Cost, ObjectPosition
from ....object_base import EventCardBase
from .....utils.desc_registry import DescDictType


class FlickeringFourLeafSigilStatus_4_3(CharactorStatusBase):
    name: Literal["Flickering Four-Leaf Sigil"] = "Flickering Four-Leaf Sigil"
    version: Literal["4.3"] = "4.3"
    usage: int = 0
    max_usage: int = 0
    icon_type: IconType = IconType.SPECIAL

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Match
    ) -> List[SwitchCharactorAction]:
        """
        At the end of every round, switch to this charactor.
        """
        current_active = match.player_tables[
            self.position.player_idx].active_charactor_idx
        if current_active == self.position.charactor_idx:
            # already active, do nothing
            return []
        return [SwitchCharactorAction(
            player_idx = self.position.player_idx,
            charactor_idx = self.position.charactor_idx
        )]


class FlickeringFourLeafSigil_4_3(EventCardBase):
    name: Literal["Flickering Four-Leaf Sigil"] = "Flickering Four-Leaf Sigil"
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost()

    def get_targets(self, match: Match) -> List[ObjectPosition]:
        # all alive charactors
        ret: List[ObjectPosition] = []
        charactors = match.player_tables[self.position.player_idx].charactors
        for charactor in charactors:
            if charactor.is_alive:
                ret.append(charactor.position)
        return ret

    def get_actions(
        self, target: ObjectPosition | None, match: Match
    ) -> List[CreateObjectAction]:
        assert target is not None
        return [CreateObjectAction(
            object_name = self.name,
            object_position = target.set_area(
                ObjectPositionType.CHARACTOR_STATUS),
            object_arguments = {}
        )]


desc: Dict[str, DescDictType] = {
    "CHARACTOR_STATUS/Flickering Four-Leaf Sigil": {
        "names": {
            "en-US": "Flickering Four-Leaf Sigil",
            "zh-CN": "浮烁的四叶印"
        },
        "descs": {
            "4.3": {
                "en-US": "At the End Phase of Every Round, you will switch to this character.",  # noqa: E501
                "zh-CN": "每个回合的结束阶段，我方都切换到此角色。"
            }
        },
    },
    "CARD/Flickering Four-Leaf Sigil": {
        "names": {
            "en-US": "Flickering Four-Leaf Sigil",
            "zh-CN": "浮烁的四叶印"
        },
        "descs": {
            "4.3": {
                "en-US": "Attached Four-Leaf Sigil to target character: At the End Phase of Every Round, you will switch to this character.",  # noqa: E501
                "zh-CN": "目标角色附属四叶印：每个回合的结束阶段，我方都切换到此角色。"
            }
        },
        "image_path": "cardface/Event_Event_Siyeyin.png",  # noqa: E501
        "id": 332027
    },
}


register_class(
    FlickeringFourLeafSigil_4_3 | FlickeringFourLeafSigilStatus_4_3, desc
)
